# Copyright (C) 2021-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all objects in a swh-storage instance and recomputes their checksums."""

import collections
import contextlib
from itertools import count, islice
import json
import logging
from typing import Iterable, Optional, Tuple, Union

import psycopg2
import tenacity

from swh.core.statsd import Statsd
from swh.journal.serializers import value_to_kafka
from swh.model import swhids
from swh.model.model import (
    Content,
    Directory,
    ObjectType,
    Release,
    Revision,
    Snapshot,
    TargetType,
)
from swh.storage.algos.directory import (
    directory_get_many_with_possibly_duplicated_entries,
)
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.storage.cassandra.storage import CassandraStorage
from swh.storage.interface import StorageInterface
from swh.storage.postgresql.storage import Storage as PostgresqlStorage

from .db import ConfigEntry, Datastore, ScrubberDb

logger = logging.getLogger(__name__)

ScrubbableObject = Union[Revision, Release, Snapshot, Directory, Content]


@contextlib.contextmanager
def postgresql_storage_db(storage):
    db = storage.get_db()
    try:
        yield db
    finally:
        storage.put_db(db)


def _get_inclusive_range_swhids(
    inclusive_range_start: Optional[bytes],
    exclusive_range_end: Optional[bytes],
    object_type: swhids.ObjectType,
) -> Tuple[swhids.CoreSWHID, swhids.CoreSWHID]:
    r"""
    Given a ``[range_start, range_end)`` right-open interval of id prefixes
    and an object type (as returned by :const:`swh.storage.backfill.RANGE_GENERATORS`),
    returns a ``[range_start_swhid, range_end_swhid]`` closed interval of SWHIDs
    suitable for the scrubber database.

    >>> _get_inclusive_range_swhids(b"\x42", None, swhids.ObjectType.SNAPSHOT)
    (CoreSWHID.from_string('swh:1:snp:4200000000000000000000000000000000000000'), CoreSWHID.from_string('swh:1:snp:ffffffffffffffffffffffffffffffffffffffff'))

    >>> _get_inclusive_range_swhids(b"\x00", b"\x12\x34", swhids.ObjectType.REVISION)
    (CoreSWHID.from_string('swh:1:rev:0000000000000000000000000000000000000000'), CoreSWHID.from_string('swh:1:rev:1233ffffffffffffffffffffffffffffffffffff'))

    """  # noqa
    range_start_swhid = swhids.CoreSWHID(
        object_type=object_type,
        object_id=(inclusive_range_start or b"").ljust(20, b"\00"),
    )
    if exclusive_range_end is None:
        inclusive_range_end = b"\xff" * 20
    else:
        # convert "1230000000..." to "122fffffff..."
        inclusive_range_end = (
            int.from_bytes(exclusive_range_end.ljust(20, b"\x00"), "big") - 1
        ).to_bytes(20, "big")
    range_end_swhid = swhids.CoreSWHID(
        object_type=object_type,
        object_id=inclusive_range_end,
    )

    return (range_start_swhid, range_end_swhid)


def get_datastore(storage) -> Datastore:
    if isinstance(storage, PostgresqlStorage):
        with postgresql_storage_db(storage) as db:
            datastore = Datastore(
                package="storage",
                cls="postgresql",
                instance=db.conn.dsn,
            )
    elif isinstance(storage, CassandraStorage):
        datastore = Datastore(
            package="storage",
            cls="cassandra",
            instance=json.dumps(
                {
                    "keyspace": storage.keyspace,
                    "hosts": storage.hosts,
                    "port": storage.port,
                }
            ),
        )
    else:
        raise NotImplementedError(f"StorageChecker(storage={storage!r}).datastore()")
    return datastore


class StorageChecker:
    """Reads a chunk of a swh-storage database, recomputes checksums, and
    reports errors in a separate database."""

    def __init__(
        self, db: ScrubberDb, config_id: int, storage: StorageInterface, limit: int = 0
    ):
        self.db = db
        self.storage = storage
        self.config_id = config_id
        self.limit = limit

        self._config: Optional[ConfigEntry] = None
        self._statsd: Optional[Statsd] = None

    @property
    def config(self) -> ConfigEntry:
        if self._config is None:
            self._config = self.db.config_get(self.config_id)

        assert self._config is not None
        return self._config

    @property
    def nb_partitions(self) -> int:
        return self.config.nb_partitions

    @property
    def object_type(self) -> swhids.ObjectType:
        return self.config.object_type

    @property
    def check_hashes(self) -> bool:
        return self.config.check_hashes

    @property
    def check_references(self) -> bool:
        return self.config.check_references

    @property
    def datastore(self) -> Datastore:
        """Returns a :class:`Datastore` instance representing the swh-storage instance
        being checked."""
        return self.config.datastore

    @property
    def statsd(self) -> Statsd:
        if self._statsd is None:
            self._statsd = Statsd(
                namespace="swh_scrubber",
                constant_tags={
                    "object_type": self.object_type,
                    "nb_partitions": self.nb_partitions,
                    "datastore_package": self.datastore.package,
                    "datastore_cls": self.datastore.cls,
                },
            )
        return self._statsd

    def run(self) -> None:
        """Runs on all objects of ``object_type`` in a partition between
        ``start_partition_id`` (inclusive) and ``end_partition_id`` (exclusive)
        """
        counter: Iterable[int] = count()
        if self.limit:
            counter = islice(counter, 0, self.limit)
        for i, partition_id in zip(
            counter, self.db.checked_partition_iter_next(self.config_id)
        ):
            logger.debug(
                "Processing %s partition %d/%d",
                self.object_type,
                partition_id,
                self.nb_partitions,
            )

            self._check_partition(self.object_type, partition_id)

            self.db.checked_partition_upsert(
                self.config_id,
                partition_id,
            )

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(psycopg2.OperationalError),
        wait=tenacity.wait_random_exponential(min=10, max=180),
    )
    def _check_partition(
        self, object_type: swhids.ObjectType, partition_id: int
    ) -> None:
        page_token = None
        while True:
            if object_type in (swhids.ObjectType.RELEASE, swhids.ObjectType.REVISION):
                method = getattr(
                    self.storage, f"{self.object_type.name.lower()}_get_partition"
                )
                page = method(partition_id, self.nb_partitions, page_token=page_token)
                objects = page.results
            elif object_type == swhids.ObjectType.DIRECTORY:
                page = self.storage.directory_get_id_partition(
                    partition_id, self.nb_partitions, page_token=page_token
                )
                directory_ids = page.results
                objects = []
                for (dir_id, item) in zip(
                    directory_ids,
                    directory_get_many_with_possibly_duplicated_entries(
                        self.storage, directory_ids
                    ),
                ):
                    assert item is not None, f"Directory {dir_id.hex()} disappeared"
                    (has_duplicate_entries, object_) = item
                    if has_duplicate_entries:
                        self.statsd.increment("duplicate_directory_entries_total")
                        self.db.corrupt_object_add(
                            object_.swhid(),
                            self.config,
                            value_to_kafka(object_.to_dict()),
                        )
                    objects.append(object_)
            elif object_type == swhids.ObjectType.SNAPSHOT:
                page = self.storage.snapshot_get_id_partition(
                    partition_id, self.nb_partitions, page_token=page_token
                )
                snapshot_ids = page.results
                objects = [
                    snapshot_get_all_branches(self.storage, snapshot_id)
                    for snapshot_id in snapshot_ids
                ]
            else:
                assert False, f"Unexpected object type: {object_type}"

            if self.check_hashes:
                with self.statsd.timed(
                    "batch_duration_seconds", tags={"operation": "check_hashes"}
                ):
                    logger.debug(
                        "Checking %s %s object hashes", len(objects), object_type
                    )
                    self.check_object_hashes(objects)
            if self.check_references:
                with self.statsd.timed(
                    "batch_duration_seconds", tags={"operation": "check_references"}
                ):
                    logger.debug(
                        "Checking %s %s object references", len(objects), object_type
                    )
                    self.check_object_references(objects)

            page_token = page.next_page_token
            if page_token is None:
                break

    def check_object_hashes(self, objects: Iterable[ScrubbableObject]):
        """Recomputes hashes, and reports mismatches."""
        count = 0
        for object_ in objects:
            if isinstance(object_, Content):
                # TODO
                continue
            real_id = object_.compute_hash()
            count += 1
            if object_.id != real_id:
                self.statsd.increment("hash_mismatch_total")
                self.db.corrupt_object_add(
                    object_.swhid(),
                    self.config,
                    value_to_kafka(object_.to_dict()),
                )
        if count:
            self.statsd.increment("objects_hashed_total", count)

    def check_object_references(self, objects: Iterable[ScrubbableObject]):
        """Check all objects references by these objects exist."""
        cnt_references = collections.defaultdict(set)
        dir_references = collections.defaultdict(set)
        rev_references = collections.defaultdict(set)
        rel_references = collections.defaultdict(set)
        snp_references = collections.defaultdict(set)

        for object_ in objects:
            swhid = object_.swhid()

            if isinstance(object_, Content):
                pass
            elif isinstance(object_, Directory):
                for entry in object_.entries:
                    if entry.type == "file":
                        cnt_references[entry.target].add(swhid)
                    elif entry.type == "dir":
                        dir_references[entry.target].add(swhid)
                    elif entry.type == "rev":
                        # dir->rev holes are not considered a problem because they
                        # happen whenever git submodules point to repositories that
                        # were not loaded yet; ignore them
                        pass
                    else:
                        assert False, entry
            elif isinstance(object_, Revision):
                dir_references[object_.directory].add(swhid)
                for parent in object_.parents:
                    rev_references[parent].add(swhid)
            elif isinstance(object_, Release):
                if object_.target is None:
                    pass
                elif object_.target_type == ObjectType.CONTENT:
                    cnt_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.DIRECTORY:
                    dir_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.REVISION:
                    rev_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.RELEASE:
                    rel_references[object_.target].add(swhid)
                else:
                    assert False, object_
            elif isinstance(object_, Snapshot):
                for branch in object_.branches.values():
                    if branch is None:
                        pass
                    elif branch.target_type == TargetType.ALIAS:
                        pass
                    elif branch.target_type == TargetType.CONTENT:
                        cnt_references[branch.target].add(swhid)
                    elif branch.target_type == TargetType.DIRECTORY:
                        dir_references[branch.target].add(swhid)
                    elif branch.target_type == TargetType.REVISION:
                        rev_references[branch.target].add(swhid)
                    elif branch.target_type == TargetType.RELEASE:
                        rel_references[branch.target].add(swhid)
                    elif branch.target_type == TargetType.SNAPSHOT:
                        snp_references[branch.target].add(swhid)
                    else:
                        assert False, (str(object_.swhid()), branch)
            else:
                assert False, object_.swhid()

        missing_cnts = set(
            self.storage.content_missing_per_sha1_git(list(cnt_references))
        )
        missing_dirs = set(self.storage.directory_missing(list(dir_references)))
        missing_revs = set(self.storage.revision_missing(list(rev_references)))
        missing_rels = set(self.storage.release_missing(list(rel_references)))
        missing_snps = set(self.storage.snapshot_missing(list(snp_references)))

        self.statsd.increment(
            "missing_object_total",
            len(missing_cnts),
            tags={"target_object_type": "content"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_dirs),
            tags={"target_object_type": "directory"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_revs),
            tags={"target_object_type": "revision"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_rels),
            tags={"target_object_type": "release"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_snps),
            tags={"target_object_type": "snapshot"},
        )

        for missing_id in missing_cnts:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.CONTENT, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, cnt_references[missing_id], self.config
            )

        for missing_id in missing_dirs:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.DIRECTORY, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, dir_references[missing_id], self.config
            )

        for missing_id in missing_revs:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.REVISION, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, rev_references[missing_id], self.config
            )

        for missing_id in missing_rels:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.RELEASE, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, rel_references[missing_id], self.config
            )

        for missing_id in missing_snps:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.SNAPSHOT, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, snp_references[missing_id], self.config
            )



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Optional
import warnings

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group
from swh.model.swhids import ObjectType


@swh_cli_group.group(name="scrubber", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.pass_context
def scrubber_cli_group(ctx, config_file: Optional[str]) -> None:
    """main command group of the datastore scrubber

    Expected config format::

        scrubber:
            cls: postgresql
            db: "service=..."    # libpq DSN

        # for storage checkers + origin locator only:
        storage:
            cls: postgresql     # cannot be remote for checkers, as they need direct
                                # access to the pg DB
            db": "service=..."  # libpq DSN
            objstorage:
                cls: memory

        # for journal checkers only:
        journal:
            # see https://docs.softwareheritage.org/devel/apidoc/swh.journal.client.html
            # for the full list of options
            sasl.mechanism: SCRAM-SHA-512
            security.protocol: SASL_SSL
            sasl.username: ...
            sasl.password: ...
            group_id: ...
            privileged: True
            message.max.bytes: 524288000
            brokers:
              - "broker1.journal.softwareheritage.org:9093
              - "broker2.journal.softwareheritage.org:9093
              - "broker3.journal.softwareheritage.org:9093
              - "broker4.journal.softwareheritage.org:9093
              - "broker5.journal.softwareheritage.org:9093
            object_types: [directory, revision, snapshot, release]
            auto_offset_reset: earliest
    """
    from swh.core import config

    from . import get_scrubber_db

    if not config_file:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")

    if config_file:
        if not os.path.exists(config_file):
            raise ValueError("%s does not exist" % config_file)
        conf = config.read(config_file)
    else:
        conf = {}

    ctx.ensure_object(dict)
    ctx.obj["config"] = conf
    if "scrubber_db" in conf:
        warnings.warn(
            "the 'scrubber_db' configuration section has been renamed to 'scrubber'; "
            f"please update your configuration file {config_file}",
            DeprecationWarning,
        )
        conf["scrubber"] = conf.pop("scrubber_db")

    if "scrubber" not in conf:
        click.echo(
            "WARNING: You must have a scrubber configured in your config file.\n"
        )
    else:
        ctx.obj["db"] = get_scrubber_db(**conf["scrubber"])


@scrubber_cli_group.group(name="check")
@click.pass_context
def scrubber_check_cli_group(ctx):
    """group of commands which read from data stores and report errors."""
    pass


@scrubber_check_cli_group.command(name="init")
@click.argument("backend", type=click.Choice(["storage", "journal"]))
@click.option(
    "--object-type",
    type=click.Choice(
        # use a hardcoded list to prevent having to load the
        # replay module at cli loading time
        [
            "snapshot",
            "revision",
            "release",
            "directory",
            # TODO:
            # "raw_extrinsic_metadata",
            # "extid",
        ]
    ),
)
@click.option("--nb-partitions", default=4096, type=int)
@click.option("--name", default=None, type=str)
@click.option("--check-hashes/--no-check-hashes", default=True)
@click.option("--check-references/--no-check-references", default=None)
@click.pass_context
def scrubber_check_init(
    ctx,
    backend: str,
    object_type: str,
    nb_partitions: int,
    name: Optional[str],
    check_hashes: bool,
    check_references: Optional[bool],
):
    """Initialise a scrubber check configuration for the datastore defined in the
    configuration file and given object_type.

    A checker configuration configuration consists simply in a set of:

    - backend: the datastore type being scrubbed (storage or journal),

    - object-type: the type of object being checked,

    - nb-pertitions: the number of partitions the hash space is divided
      in; must be a power of 2,

    - name: an unique name for easier reference,

    - check-hashes: flag (default to True) to select the hash validation step for
      this scrubbing configuration,

    - check-references: flag (default to True for storage and False for the journal
      backend) to select the reference validation step for this scrubbing configuration.
    """
    if not object_type or not name:
        raise click.ClickException(
            "Invalid parameters: you must provide the object type and configuration name"
        )

    conf = ctx.obj["config"]
    db = ctx.obj["db"]

    if backend == "storage":
        if check_references is None:
            check_references = True
        if "storage" not in conf:
            raise click.ClickException(
                "You must have a storage configured in your config file."
            )
        from swh.storage import get_storage

        from .storage_checker import get_datastore as get_storage_datastore

        datastore = get_storage_datastore(storage=get_storage(**conf["storage"]))
        db.datastore_get_or_add(datastore)
    elif backend == "journal":
        if check_references is None:
            check_references = False
        if "journal" not in conf:
            raise click.ClickException(
                "You must have a journal configured in your config file."
            )
        from .journal_checker import get_datastore as get_journal_datastore

        datastore = get_journal_datastore(journal_cfg=conf["journal"])
        db.datastore_get_or_add(datastore)
    else:
        raise click.ClickException(f"Backend type {backend} is not supported")

    if db.config_get_by_name(name):
        raise click.ClickException(f"Configuration {name} already exists")

    assert check_references is not None

    config_id = db.config_add(
        name,
        datastore,
        getattr(ObjectType, object_type.upper()),
        nb_partitions,
        check_hashes=check_hashes,
        check_references=check_references,
    )
    click.echo(
        f"Created configuration {name} [{config_id}] for checking {object_type} "
        f"in {datastore.cls} {datastore.package}"
    )


@scrubber_check_cli_group.command(name="list")
@click.pass_context
def scrubber_check_list(
    ctx,
):
    """List the know configurations"""
    conf = ctx.obj["config"]
    if "storage" not in conf:
        ctx.fail("You must have a storage configured in your config file.")

    db = ctx.obj["db"]

    for id_, cfg in db.config_iter():
        ds = cfg.datastore
        if not ds:
            click.echo(
                f"[{id_}] {cfg.name}: Invalid configuration entry; datastore not found"
            )
        else:
            click.echo(
                f"[{id_}] {cfg.name}: {cfg.object_type}, {cfg.nb_partitions}, "
                f"{ds.package}:{ds.cls} ({ds.instance})"
            )


@scrubber_check_cli_group.command(name="stalled")
@click.argument(
    "name",
    type=str,
    default=None,
    required=False,  # can be given by config_id instead
)
@click.option(
    "--config-id",
    type=int,
)
@click.option(
    "--for",
    "delay",
    type=str,
    default="auto",
    help="Delay for a partition to be considered as stuck; in seconds or 'auto'",
)
@click.option(
    "--reset",
    is_flag=True,
    default=False,
    help="Reset the stalled partition so it can be grabbed by a scrubber worker",
)
@click.pass_context
def scrubber_check_stalled(
    ctx, name: str, config_id: int, delay: Optional[str], reset: bool
):
    """List the stuck partitions for a given config"""
    import datetime

    from humanize import naturaldate, naturaldelta

    db = ctx.obj["db"]
    if name and config_id is None:
        config_id = db.config_get_by_name(name)

    if config_id is None:
        raise click.ClickException("A valid configuration name/id must be set")

    cfg = db.config_get(config_id)
    delay_td: Optional[datetime.timedelta]
    if delay == "auto":
        delay_td = None
    elif delay:
        delay_td = datetime.timedelta(seconds=int(delay))
    in_flight = list(db.checked_partition_get_stuck(config_id, delay_td))
    if in_flight:
        click.echo(
            f"Stuck partitions for {cfg.name} [id={config_id}, "
            f"type={cfg.object_type.name.lower()}]:"
        )
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        for partition, stuck_since in in_flight:
            click.echo(
                f"{partition}:\tstuck since {naturaldate(stuck_since)} "
                f"({naturaldelta(now-stuck_since)})"
            )
            if reset:
                if db.checked_partition_reset(config_id, partition):
                    click.echo("\tpartition reset")
                else:
                    click.echo("\tpartition NOT reset")

    else:
        click.echo(
            f"No stuck partition found for {cfg.name} [id={config_id}, "
            f"type={cfg.object_type.name.lower()}]"
        )


@scrubber_check_cli_group.command(name="storage")
@click.argument(
    "name",
    type=str,
    default=None,
    required=False,  # can be given by config_id instead
)
@click.option(
    "--config-id",
    type=int,
    help="Config ID (is config name is not given as argument)",
)
@click.option("--limit", default=0, type=int)
@click.pass_context
def scrubber_check_storage(
    ctx,
    name: str,
    config_id: int,
    limit: int,
):
    """Reads a swh-storage instance, and reports corrupt objects to the scrubber DB.

    This runs a single thread; parallelism is achieved by running this command
    multiple times.

    This command references an existing scrubbing configuration (either by name
    or by id); the configuration holds the object type, number of partitions
    and the storage configuration this scrubbing session will check on.

    All objects of type ``object_type`` are ordered, and split into the given
    number of partitions.

    Then, this process will check all partitions. The status of the ongoing
    check session is stored in the database, so the number of concurrent
    workers can be dynamically adjusted.

    """  # noqa
    conf = ctx.obj["config"]
    if "storage" not in conf:
        ctx.fail("You must have a storage configured in your config file.")
    db = ctx.obj["db"]

    from swh.storage import get_storage

    from .storage_checker import StorageChecker

    if name and config_id is None:
        from .storage_checker import get_datastore as get_storage_datastore

        cfg = conf["storage"]
        datastore = get_storage_datastore(storage=get_storage(**cfg))
        datastore_id = db.datastore_get_or_add(datastore)
        config_id = db.config_get_by_name(name, datastore_id)
    elif name is None and config_id is not None:
        assert db.config_get(config_id) is not None

    if config_id is None:
        raise click.ClickException("A valid configuration name/id must be set")

    checker = StorageChecker(
        db=ctx.obj["db"],
        storage=get_storage(**conf["storage"]),
        config_id=config_id,
        limit=limit,
    )

    checker.run()


@scrubber_check_cli_group.command(name="journal")
@click.argument(
    "name",
    type=str,
    default=None,
    required=False,  # can be given by config_id instead
)
@click.option(
    "--config-id",
    type=int,
    help="Config ID (is config name is not given as argument)",
)
@click.pass_context
def scrubber_check_journal(ctx, name, config_id) -> None:
    """Reads a complete kafka journal, and reports corrupt objects to
    the scrubber DB."""
    conf = ctx.obj["config"]
    if "journal" not in conf:
        ctx.fail("You must have a journal configured in your config file.")
    db = ctx.obj["db"]

    if name and config_id is None:
        from .journal_checker import get_datastore as get_journal_datastore

        cfg = conf["journal"]
        datastore = get_journal_datastore(journal_cfg=cfg)
        datastore_id = db.datastore_get_or_add(datastore)
        config_id = db.config_get_by_name(name, datastore_id)
    elif name is None and config_id is not None:
        assert db.config_get(config_id) is not None

    if config_id is None:
        raise click.ClickException("A valid configuration name/id must be set")

    from .journal_checker import JournalChecker

    checker = JournalChecker(
        db=ctx.obj["db"],
        journal=conf["journal"],
        config_id=config_id,
    )

    checker.run()


@scrubber_cli_group.command(name="locate")
@click.option("--start-object", default="swh:1:cnt:" + "00" * 20)
@click.option("--end-object", default="swh:1:snp:" + "ff" * 20)
@click.pass_context
def scrubber_locate_origins(ctx, start_object: str, end_object: str):
    """For each known corrupt object reported in the scrubber DB, looks up origins
    that may contain this object, and records them; so they can be used later
    for recovery."""
    conf = ctx.obj["config"]
    if "storage" not in conf:
        ctx.fail("You must have a storage configured in your config file.")
    if "graph" not in conf:
        ctx.fail("You must have a graph configured in your config file.")

    from swh.graph.http_client import RemoteGraphClient
    from swh.model.model import CoreSWHID
    from swh.storage import get_storage

    from .origin_locator import OriginLocator

    locator = OriginLocator(
        db=ctx.obj["db"],
        storage=get_storage(**conf["storage"]),
        graph=RemoteGraphClient(**conf["graph"]),
        start_object=CoreSWHID.from_string(start_object),
        end_object=CoreSWHID.from_string(end_object),
    )

    locator.run()


@scrubber_cli_group.command(name="fix")
@click.option("--start-object", default="swh:1:cnt:" + "00" * 20)
@click.option("--end-object", default="swh:1:snp:" + "ff" * 20)
@click.pass_context
def scrubber_fix_objects(ctx, start_object: str, end_object: str):
    """For each known corrupt object reported in the scrubber DB, looks up origins
    that may contain this object, and records them; so they can be used later
    for recovery."""
    from swh.model.model import CoreSWHID

    from .fixer import Fixer

    fixer = Fixer(
        db=ctx.obj["db"],
        start_object=CoreSWHID.from_string(start_object),
        end_object=CoreSWHID.from_string(end_object),
    )

    fixer.run()



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Lists corrupt objects in the scrubber database, and lists candidate origins
to recover them from."""

import dataclasses
import itertools
import logging
from typing import Iterable, Union

import psycopg2

from swh.core.utils import grouper
from swh.graph.http_client import GraphArgumentException, RemoteGraphClient
from swh.model.model import Directory, Release, Revision, Snapshot
from swh.model.swhids import CoreSWHID, ExtendedSWHID
from swh.storage.interface import StorageInterface

from .db import CorruptObject, ScrubberDb
from .utils import iter_corrupt_objects

logger = logging.getLogger(__name__)

ScrubbableObject = Union[Revision, Release, Snapshot, Directory]


def get_origins(
    graph: RemoteGraphClient, storage: StorageInterface, swhid: CoreSWHID
) -> Iterable[str]:
    try:
        origin_swhids = [
            ExtendedSWHID.from_string(line)
            for line in graph.leaves(str(swhid), direction="backward")
            if line.startswith("swh:1:ori:")
        ]
    except GraphArgumentException:
        return

    for origin_swhid_group in grouper(origin_swhids, 10):
        origin_swhid_group = list(origin_swhid_group)
        for (origin, origin_swhid) in zip(
            storage.origin_get_by_sha1(
                [origin_swhid.object_id for origin_swhid in origin_swhid_group]
            ),
            origin_swhid_group,
        ):
            if origin is None:
                logger.error("%s found in graph but missing from storage", origin_swhid)
            else:
                yield origin["url"]


@dataclasses.dataclass
class OriginLocator:
    """Reads a chunk of corrupt objects in the swh-scrubber database, then writes
    to the same database a list of origins they might be recovered from."""

    db: ScrubberDb
    """Database to read from and write to."""
    graph: RemoteGraphClient
    storage: StorageInterface
    """Used to resolve origin SHA1s to URLs."""

    start_object: CoreSWHID
    """Minimum SWHID to check (in alphabetical order)"""
    end_object: CoreSWHID
    """Maximum SWHID to check (in alphabetical order)"""

    def run(self):
        iter_corrupt_objects(
            self.db,
            self.start_object,
            self.end_object,
            None,
            self.handle_corrupt_object,
        )

    def handle_corrupt_object(
        self, corrupt_object: CorruptObject, cur: psycopg2.extensions.cursor
    ) -> None:
        origins = get_origins(self.graph, self.storage, corrupt_object.id)

        # Keep only 100 origins, to avoid flooding the DB.
        # It is very unlikely an object disappred from 100 somwhat-randomly sampled
        # origins.
        first_origins = list(itertools.islice(origins, 0, 100))

        self.db.object_origin_add(cur, corrupt_object.id, first_origins)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging
from pathlib import Path
import subprocess
from unittest.mock import MagicMock
import zlib

import attr
import pytest

from swh.journal.serializers import kafka_to_value, value_to_kafka
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Directory, DirectoryEntry
from swh.scrubber.db import (
    ConfigEntry,
    CorruptObject,
    Datastore,
    FixedObject,
    ScrubberDb,
)
from swh.scrubber.fixer import Fixer

from .conftest import CORRUPT_DIRECTORY, ORIGINAL_DIRECTORY


@pytest.fixture
def corrupt_object(config_entry) -> CorruptObject:
    return CorruptObject(
        id=ORIGINAL_DIRECTORY.swhid(),
        config=config_entry,
        first_occurrence=datetime.datetime.now(tz=datetime.timezone.utc),
        object_=value_to_kafka(CORRUPT_DIRECTORY.to_dict()),
    )


def test_no_object(scrubber_db: ScrubberDb, mocker) -> None:
    """There is no object to recover -> nothing happens"""
    fixer = Fixer(db=scrubber_db)
    fixer.run()

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM fixed_object")
        assert cur.fetchone() == (0,)


def test_no_origin(
    scrubber_db: ScrubberDb, corrupt_object: CorruptObject, mocker
) -> None:
    """There is no origin to recover objects from -> nothing happens"""
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )

    fixer = Fixer(db=scrubber_db)
    fixer.run()

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM fixed_object")
        assert cur.fetchone() == (0,)


def test_already_fixed(
    scrubber_db: ScrubberDb, corrupt_object: CorruptObject, datastore: Datastore, mocker
) -> None:
    """All corrupt objects are already fixed -> nothing happens"""
    fixed_object = FixedObject(
        id=corrupt_object.id,
        object_=value_to_kafka(ORIGINAL_DIRECTORY.to_dict()),
        method="whatever means necessary",
    )
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )
    with scrubber_db.cursor() as cur:
        scrubber_db.object_origin_add(cur, corrupt_object.id, ["http://example.org/"])
        scrubber_db.fixed_object_add(cur, [fixed_object])

    subprocess_run = mocker.patch("subprocess.run")

    scrubber_db = MagicMock(wraps=scrubber_db)

    fixer = Fixer(db=scrubber_db)
    fixer.run()

    # Check the Fixer did not try to fix the object again
    scrubber_db.fixed_object_add.assert_not_called()
    subprocess_run.assert_not_called()
    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT id, method FROM fixed_object")
        assert list(cur) == [(str(fixed_object.id), fixed_object.method)]


def _run_fixer_with_clone(
    scrubber_db: ScrubberDb,
    mocker,
    caplog,
    corrupt_object: CorruptObject,
    subprocess_run_side_effect,
) -> None:
    """Helper for all tests that involve running the fixer with a clone:
    adds a corrupt object and an origin to the DB, mocks subprocess.run with the
    given function, and runs the fixer with caplog"""
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )
    with scrubber_db.cursor() as cur:
        scrubber_db.object_origin_add(cur, corrupt_object.id, ["http://example.org/"])

    subprocess_run = mocker.patch(
        "subprocess.run", side_effect=subprocess_run_side_effect
    )

    fixer = Fixer(db=scrubber_db)
    with caplog.at_level(logging.CRITICAL):
        with caplog.at_level(logging.INFO, logger="swh.scrubber.fixer"):
            fixer.run()

    subprocess_run.assert_called()


def test_failed_clone(
    scrubber_db: ScrubberDb, corrupt_object: CorruptObject, mocker, caplog
) -> None:
    """Corrupt object found with an origin, but the origin's clone is broken somehow"""
    scrubber_db = MagicMock(wraps=scrubber_db)

    _run_fixer_with_clone(
        scrubber_db,
        mocker,
        caplog,
        corrupt_object=corrupt_object,
        subprocess_run_side_effect=subprocess.CalledProcessError(1, "foo"),
    )

    scrubber_db.fixed_object_add.assert_not_called()
    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT id, method FROM fixed_object")
        assert list(cur) == []

    assert (
        "swh.scrubber.fixer",
        logging.ERROR,
        "Failed to clone http://example.org/",
    ) in caplog.record_tuples


def test_empty_origin(
    scrubber_db: ScrubberDb, corrupt_object: CorruptObject, mocker, caplog
) -> None:
    """Corrupt object found with an origin, but the origin's clone is missing
    the object"""
    scrubber_db = MagicMock(wraps=scrubber_db)
    real_subprocess_run = subprocess.run

    def subprocess_run(args, **kwargs):
        (*head, path) = args
        assert head == ["git", "clone", "--bare", "http://example.org/"]
        real_subprocess_run(["git", "init", "--bare", path])

    _run_fixer_with_clone(
        scrubber_db,
        mocker,
        caplog,
        corrupt_object=corrupt_object,
        subprocess_run_side_effect=subprocess_run,
    )

    scrubber_db.fixed_object_add.assert_not_called()
    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT id, method FROM fixed_object")
        assert list(cur) == []

    assert (
        "swh.scrubber.fixer",
        logging.INFO,
        "swh:1:dir:61992617462fff81509bda4a24b54c96ea74a007 not found in origin",
    ) in caplog.record_tuples


def test_parseable_directory_from_origin(
    scrubber_db: ScrubberDb, corrupt_object: CorruptObject, mocker, caplog
) -> None:
    """Corrupt object found with an origin, and the object is found in the origin's
    clone as expected."""
    scrubber_db = MagicMock(wraps=scrubber_db)
    real_subprocess_run = subprocess.run

    def subprocess_run(args, **kwargs):
        (*head, path) = args
        assert head == ["git", "clone", "--bare", "http://example.org/"]
        real_subprocess_run(["git", "init", "--bare", path])
        object_dir_path = Path(path) / "objects/61"
        object_path = object_dir_path / "992617462fff81509bda4a24b54c96ea74a007"
        object_dir_path.mkdir()
        with open(object_path, "wb") as fd:
            fd.write(zlib.compress(ORIGINAL_DIRECTORY.raw_manifest))

    _run_fixer_with_clone(
        scrubber_db,
        mocker,
        caplog,
        corrupt_object=corrupt_object,
        subprocess_run_side_effect=subprocess_run,
    )

    scrubber_db.fixed_object_add.assert_called_once()
    fixed_objects = list(scrubber_db.fixed_object_iter())
    assert len(fixed_objects) == 1

    assert fixed_objects[0].id == ORIGINAL_DIRECTORY.swhid()
    assert fixed_objects[0].method == "from_origin"
    assert (
        Directory.from_dict(kafka_to_value(fixed_objects[0].object_))
        == ORIGINAL_DIRECTORY
    )

    assert caplog.record_tuples == []


def test_unparseable_directory(
    scrubber_db: ScrubberDb, config_entry: ConfigEntry, mocker, caplog
) -> None:
    """Corrupt object found with an origin, and the object is found in the origin's
    clone as expected; but Dulwich cannot parse it.
    It was probably loaded by an old version of the loader that was more permissive,
    by using libgit2."""
    scrubber_db = MagicMock(wraps=scrubber_db)
    real_subprocess_run = subprocess.run

    raw_manifest = b"this is not a parseable manifest"
    raw_manifest = f"tree {len(raw_manifest)}\x00".encode() + raw_manifest

    original_directory = Directory(
        entries=(
            DirectoryEntry(
                name=b"dir1",
                type="dir",
                target=hash_to_bytes("4b825dc642cb6eb9a060e54bf8d69288fbee4904"),
                perms=0o040755,
            ),
        ),
        raw_manifest=raw_manifest,
    )
    assert original_directory.id.hex() == "a518fa6b46bad74e95588d2bfdf4455398a2216a"

    corrupt_directory = attr.evolve(original_directory, raw_manifest=None)
    corrupt_object = CorruptObject(
        id=original_directory.swhid(),
        config=config_entry,
        object_=value_to_kafka(corrupt_directory.to_dict()),
        first_occurrence=datetime.datetime.now(tz=datetime.timezone.utc),
    )

    def subprocess_run(args, **kwargs):
        (*head, path) = args
        if head[0:2] != ["git", "clone"]:
            return real_subprocess_run(args, **kwargs)
        assert head == ["git", "clone", "--bare", "http://example.org/"]
        real_subprocess_run(["git", "init", "--bare", path])
        object_dir_path = Path(path) / "objects/a5"
        object_path = object_dir_path / "18fa6b46bad74e95588d2bfdf4455398a2216a"
        object_dir_path.mkdir()
        with open(object_path, "wb") as fd:
            fd.write(zlib.compress(raw_manifest))

    _run_fixer_with_clone(
        scrubber_db,
        mocker,
        caplog,
        corrupt_object=corrupt_object,
        subprocess_run_side_effect=subprocess_run,
    )

    scrubber_db.fixed_object_add.assert_called_once()
    fixed_objects = list(scrubber_db.fixed_object_iter())
    assert len(fixed_objects) == 1

    assert fixed_objects[0].id == original_directory.swhid()
    assert fixed_objects[0].method == "manifest_from_origin"
    assert (
        Directory.from_dict(kafka_to_value(fixed_objects[0].object_))
        == original_directory
    )

    assert caplog.record_tuples == [
        (
            "swh.scrubber.fixer",
            logging.INFO,
            r"Dulwich failed to parse b'tree 32\x00this is not a parseable manifest'",
        )
    ]



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from swh.scrubber.db import ScrubberDb


def get_scrubber_db(cls: str, **kwargs) -> ScrubberDb:
    if cls not in ("local", "postgresql"):
        raise ValueError(
            f"Unknown scrubber db class '{cls}', use 'postgresql' instead."
        )

    from swh.scrubber.db import ScrubberDb

    return ScrubberDb.connect(kwargs.pop("db"), **kwargs)


get_datastore = get_scrubber_db



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tempfile
from unittest.mock import MagicMock, call

from click.testing import CliRunner
import yaml

from swh.model.swhids import CoreSWHID
from swh.scrubber.cli import scrubber_cli_group
from swh.scrubber.storage_checker import postgresql_storage_db


def invoke(
    scrubber_db,
    args,
    storage=None,
    kafka_server=None,
    kafka_prefix=None,
    kafka_consumer_group=None,
):
    runner = CliRunner()

    config = {
        "scrubber": {"cls": "postgresql", "db": scrubber_db.conn.dsn},
        "graph": {"url": "http://graph.example.org:5009/"},
    }
    if storage:
        with postgresql_storage_db(storage) as db:
            config["storage"] = {
                "cls": "postgresql",
                "db": db.conn.dsn,
                "objstorage": {"cls": "memory"},
            }

    assert (
        (kafka_server is None)
        == (kafka_prefix is None)
        == (kafka_consumer_group is None)
    )
    if kafka_server:
        config["journal"] = dict(
            cls="kafka",
            brokers=kafka_server,
            group_id=kafka_consumer_group,
            prefix=kafka_prefix,
            stop_on_eof=True,
        )

    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        yaml.dump(config, config_fd)
        config_fd.seek(0)
        args = ["-C" + config_fd.name] + list(args)
        result = runner.invoke(scrubber_cli_group, args, catch_exceptions=False)
    return result


def test_help_main(mocker, scrubber_db, swh_storage):
    result = invoke(
        scrubber_db,
        [
            "--help",
        ],
    )
    assert result.exit_code == 0, result.output
    output = result.output.splitlines(keepends=False)
    msg = "Usage: scrubber [OPTIONS] COMMAND [ARGS]..."
    assert output[0] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    assert commands == ["check", "fix", "locate"]


def test_help_check(mocker, scrubber_db, swh_storage):
    result = invoke(
        scrubber_db,
        [
            "check",
            "--help",
        ],
    )
    assert result.exit_code == 0, result.output
    output = result.output.splitlines(keepends=False)
    msg = "Usage: scrubber check [OPTIONS] COMMAND [ARGS]..."
    assert output[0] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    # With older click version (e.g. 7.0-1), the text wrapping can be different,
    # resulting in some docstring text included in this command list, so check we find
    # the expected commands instead
    for command in ["init", "journal", "list", "stalled", "storage"]:
        assert command in commands

    # without a config file, --help should still work but with an extra message
    result = CliRunner().invoke(
        scrubber_cli_group, ["check", "--help"], catch_exceptions=False
    )
    output = result.output.splitlines(keepends=False)
    msg = "WARNING: You must have a scrubber configured in your config file."
    assert output[0] == msg
    msg = "Usage: scrubber check [OPTIONS] COMMAND [ARGS]..."
    assert output[2] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    # With older click version (e.g. 7.0-1), the text wrapping can be different,
    # resulting in some docstring text included in this command list, so check we find
    # the expected commands instead
    for command in ["init", "journal", "list", "stalled", "storage"]:
        assert command in commands


def test_check_init(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output
    msg = "Created configuration cfg1 [1] for checking snapshot in postgresql storage"
    assert result.output.strip() == msg

    # error: cfg name already exists
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "8",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 1, result.output
    msg = "Error: Configuration cfg1 already exists"
    assert result.output.strip() == msg


def test_check_init_storage_flags(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    arg_list = [
        "check",
        "init",
        "storage",
        "--object-type",
        "snapshot",
        "--nb-partitions",
        "4",
        "--name",
    ]

    name = "cfg1"
    result = invoke(
        scrubber_db,
        arg_list + [name],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is True

    name = "cfg2"
    result = invoke(
        scrubber_db,
        arg_list + [name, "--no-check-references"],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is False

    name = "cfg3"
    result = invoke(
        scrubber_db,
        arg_list + [name, "--no-check-hashes"],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is False
    assert cfg_entry.check_references is True


def test_check_init_journal_flags(
    mocker, scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    arg_list = [
        "check",
        "init",
        "journal",
        "--object-type",
        "snapshot",
        "--name",
    ]

    name = "cfg1"
    result = invoke(
        scrubber_db,
        arg_list + [name],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert result.exit_code == 0, result.output

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is False


def test_check_storage(mocker, scrubber_db, swh_storage):
    storage_checker = MagicMock()
    StorageChecker = mocker.patch(
        "swh.scrubber.storage_checker.StorageChecker", return_value=storage_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output
    msg = "Created configuration cfg1 [1] for checking snapshot in postgresql storage"
    assert result.output.strip() == msg

    result = invoke(scrubber_db, ["check", "storage", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    assert result.output == ""

    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)
    StorageChecker.assert_called_once_with(
        db=scrubber_db,
        config_id=1,
        storage=StorageChecker.mock_calls[0][2]["storage"],
        limit=0,
    )
    assert storage_checker.method_calls == [call.run()]

    # using the config id instead of the config name
    result = invoke(
        scrubber_db, ["check", "storage", "--config-id", "1"], storage=swh_storage
    )
    assert result.exit_code == 0, result.output
    assert result.output == ""


def test_check_storage_ko(mocker, scrubber_db, swh_storage):
    # using the config id instead of the config name
    result = invoke(scrubber_db, ["check", "storage"], storage=swh_storage)
    assert result.exit_code == 1, result.output
    assert result.output == "Error: A valid configuration name/id must be set\n"


def test_check_list(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    assert result.output == ""
    with swh_storage.db() as db:
        dsn = db.conn.dsn

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = f"[1] cfg1: snapshot, 4, storage:postgresql ({dsn})\n"
    assert result.output == expected, result.output


def test_check_stalled(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    assert result.output == ""

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a partition started 20mn ago
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 0, now() - '20m'::interval, NULL);"
        )

    # there are no existing completed partition, defaults to 1h to be considered as stalled
    # so a partition just added is not stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a partition started 2 hours from now
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 1, now() - '2h'::interval, NULL);"
        )
    # it is considered as stalled by default
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output

    # explicitly specify a delay > 2h to be considered as stelles: no one stalled
    result = invoke(
        scrubber_db, ["check", "stalled", "--for", "8000", "cfg1"], storage=swh_storage
    )
    assert result.exit_code == 0, result.output
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a transaction that took 5mn to run
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 2, now() - '2h'::interval, now() - '1h55m'::interval);"
        )
    # so now both partitions 0 and 1 should be considered a stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output


def test_check_reset(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    assert result.output == ""

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 0, result.output

    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a few partitions
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 0, now() - '20m'::interval, NULL);"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 1, now() - '2h'::interval, NULL);"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 2, now() - '2h'::interval, now() - '1h55m'::interval);"
        )

    # partitions 0 and 1 are considered as stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output

    # let's reset them
    result = invoke(
        scrubber_db, ["check", "stalled", "--reset", "cfg1"], storage=swh_storage
    )
    assert result.exit_code == 0, result.output
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
	partition reset
1:	stuck since today (2 hours)
	partition reset
"""  # noqa: W191,E101
    assert result.output == expected, result.output
    with scrubber_db.transaction() as cur:
        cur.execute(
            "SELECT partition_id, end_date "
            "FROM checked_partition "
            "WHERE config_id=1 AND start_date is NULL"
        )
        assert cur.fetchall() == [(0, None), (1, None)]

    # for good measure, check the next few selected partitions, expected 0, 1 and 3
    assert next(scrubber_db.checked_partition_iter_next(1)) == 0
    assert next(scrubber_db.checked_partition_iter_next(1)) == 1
    assert next(scrubber_db.checked_partition_iter_next(1)) == 3


def test_check_journal(
    mocker, scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    journal_checker = MagicMock()
    JournalChecker = mocker.patch(
        "swh.scrubber.journal_checker.JournalChecker", return_value=journal_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "journal",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert result.exit_code == 0, result.output
    msg = "Created configuration cfg1 [1] for checking snapshot in kafka journal"
    assert result.output.strip() == msg

    result = invoke(
        scrubber_db,
        ["check", "journal", "cfg1"],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert result.exit_code == 0, result.output
    assert result.output == ""

    assert get_scrubber_db.call_count == 2
    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)

    JournalChecker.assert_called_once_with(
        db=scrubber_db,
        journal={
            "brokers": kafka_server,
            "cls": "kafka",
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
            "stop_on_eof": True,
        },
        config_id=1,
    )
    assert journal_checker.method_calls == [call.run()]


def test_locate_origins(mocker, scrubber_db, swh_storage):
    origin_locator = MagicMock()
    OriginLocator = mocker.patch(
        "swh.scrubber.origin_locator.OriginLocator", return_value=origin_locator
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(scrubber_db, ["locate"], storage=swh_storage)
    assert result.exit_code == 0, result.output
    assert result.output == ""

    get_scrubber_db.assert_called_once_with(cls="postgresql", db=scrubber_db.conn.dsn)
    OriginLocator.assert_called_once_with(
        db=scrubber_db,
        storage=OriginLocator.mock_calls[0][2]["storage"],
        graph=OriginLocator.mock_calls[0][2]["graph"],
        start_object=CoreSWHID.from_string("swh:1:cnt:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:snp:" + "ff" * 20),
    )
    assert origin_locator.method_calls == [call.run()]


def test_fix_objects(mocker, scrubber_db):
    fixer = MagicMock()
    Fixer = mocker.patch("swh.scrubber.fixer.Fixer", return_value=fixer)
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(scrubber_db, ["fix"])
    assert result.exit_code == 0, result.output
    assert result.output == ""

    get_scrubber_db.assert_called_once_with(cls="postgresql", db=scrubber_db.conn.dsn)
    Fixer.assert_called_once_with(
        db=scrubber_db,
        start_object=CoreSWHID.from_string("swh:1:cnt:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:snp:" + "ff" * 20),
    )
    assert fixer.method_calls == [call.run()]



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scrubber.db import Datastore
from swh.scrubber.storage_checker import postgresql_storage_db

from .storage_checker_tests import *  # noqa


@pytest.fixture
def swh_storage_backend_config(swh_storage_postgresql_backend_config):
    return swh_storage_postgresql_backend_config


@pytest.fixture
def datastore(swh_storage):
    with postgresql_storage_db(swh_storage) as db:
        return Datastore(
            package="storage",
            cls="postgresql",
            instance=db.conn.dsn,
        )



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr
import msgpack
import pytest

from swh.journal.serializers import kafka_to_value
from swh.model import model, swhids
from swh.model.tests import swh_model_data
from swh.scrubber.storage_checker import StorageChecker

CONTENT1 = model.Content.from_data(b"foo")
DIRECTORY1 = model.Directory(
    entries=(
        model.DirectoryEntry(
            target=CONTENT1.sha1_git, type="file", name=b"file1", perms=0o1
        ),
    )
)
DIRECTORY2 = model.Directory(
    entries=(
        model.DirectoryEntry(
            target=CONTENT1.sha1_git, type="file", name=b"file2", perms=0o1
        ),
        model.DirectoryEntry(target=DIRECTORY1.id, type="dir", name=b"dir1", perms=0o1),
        model.DirectoryEntry(target=b"\x00" * 20, type="rev", name=b"rev1", perms=0o1),
    )
)
REVISION1 = model.Revision(
    message=b"blah",
    directory=DIRECTORY2.id,
    author=None,
    committer=None,
    date=None,
    committer_date=None,
    type=model.RevisionType.GIT,
    synthetic=True,
)
RELEASE1 = model.Release(
    message=b"blih",
    name=b"bluh",
    target_type=model.ObjectType.REVISION,
    target=REVISION1.id,
    synthetic=True,
)
SNAPSHOT1 = model.Snapshot(
    branches={
        b"rel1": model.SnapshotBranch(
            target_type=model.TargetType.RELEASE, target=RELEASE1.id
        ),
    }
)


EXPECTED_PARTITIONS = {
    (swhids.ObjectType.SNAPSHOT, 0, 1),
    (swhids.ObjectType.DIRECTORY, 0, 1),
    (swhids.ObjectType.REVISION, 0, 1),
    (swhids.ObjectType.RELEASE, 0, 1),
}

OBJECT_TYPES = (
    swhids.ObjectType.SNAPSHOT,
    swhids.ObjectType.DIRECTORY,
    swhids.ObjectType.REVISION,
    swhids.ObjectType.RELEASE,
)


def assert_checked_ranges(
    scrubber_db, config, expected_ranges, before_date=None, after_date=None
):
    checked_ranges = set()
    for object_type, config_id in config:
        if before_date is not None:
            assert all(
                before_date < date < after_date
                for (_, _, date, _) in scrubber_db.checked_partition_iter(config_id)
            )

        checked_ranges.update(
            {
                (object_type, partition, nb_partitions)
                for (
                    partition,
                    nb_partitions,
                    start_date,
                    end_date,
                ) in scrubber_db.checked_partition_iter(config_id)
            }
        )
    assert checked_ranges == expected_ranges


def test_no_corruption(scrubber_db, datastore, swh_storage):
    swh_storage.directory_add(swh_model_data.DIRECTORIES)
    swh_storage.revision_add(swh_model_data.REVISIONS)
    swh_storage.release_add(swh_model_data.RELEASES)
    swh_storage.snapshot_add(swh_model_data.SNAPSHOTS)

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add(
            f"cfg_{object_type.name}", datastore, object_type, 1
        )
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    assert list(scrubber_db.corrupt_object_iter()) == []

    assert_checked_ranges(
        scrubber_db, config, EXPECTED_PARTITIONS, before_date, after_date
    )


@pytest.mark.parametrize("corrupt_idx", range(len(swh_model_data.SNAPSHOTS)))
def test_corrupt_snapshot(scrubber_db, datastore, swh_storage, corrupt_idx):
    snapshots = list(swh_model_data.SNAPSHOTS)
    snapshots[corrupt_idx] = attr.evolve(snapshots[corrupt_idx], id=b"\x00" * 20)
    swh_storage.snapshot_add(snapshots)

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add(
            f"cfg_{object_type.name}", datastore, object_type, 1
        )
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == swhids.CoreSWHID.from_string(
        "swh:1:snp:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].config.datastore == datastore
    assert (
        before_date - datetime.timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + datetime.timedelta(seconds=5)
    )
    assert (
        kafka_to_value(corrupt_objects[0].object_) == snapshots[corrupt_idx].to_dict()
    )

    assert_checked_ranges(
        scrubber_db, config, EXPECTED_PARTITIONS, before_date, after_date
    )


def test_corrupt_snapshots_same_batch(scrubber_db, datastore, swh_storage):
    snapshots = list(swh_model_data.SNAPSHOTS)
    for i in (0, 1):
        snapshots[i] = attr.evolve(snapshots[i], id=bytes([i]) * 20)
    swh_storage.snapshot_add(snapshots)

    config_id = scrubber_db.config_add("cfg1", datastore, swhids.ObjectType.SNAPSHOT, 1)
    StorageChecker(
        db=scrubber_db,
        storage=swh_storage,
        config_id=config_id,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 2
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in [
            "swh:1:snp:0000000000000000000000000000000000000000",
            "swh:1:snp:0101010101010101010101010101010101010101",
        ]
    }

    assert_checked_ranges(
        scrubber_db,
        [(swhids.ObjectType.SNAPSHOT, config_id)],
        {(swhids.ObjectType.SNAPSHOT, 0, 1)},
    )


def test_corrupt_snapshots_different_batches(scrubber_db, datastore, swh_storage):
    # FIXME: this is brittle, because it relies on objects being on different
    # partitions. In particular on Cassandra, it will break if the hashing scheme
    # or hash algorithm changes.
    snapshots = list(swh_model_data.SNAPSHOTS)
    snapshots.extend(
        [attr.evolve(snapshots[0], id=bytes([17 * i]) * 20) for i in range(16)]
    )
    swh_storage.snapshot_add(snapshots)

    config_id = scrubber_db.config_add(
        "cfg1", datastore, swhids.ObjectType.SNAPSHOT, 16
    )
    StorageChecker(
        db=scrubber_db,
        storage=swh_storage,
        config_id=config_id,
        limit=8,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 8

    # Simulates resuming from a different process
    scrubber_db.datastore_get_or_add.cache_clear()

    StorageChecker(
        db=scrubber_db,
        storage=swh_storage,
        config_id=config_id,
        limit=8,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 16
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in ["swh:1:snp:" + f"{i:x}" * 40 for i in range(16)]
    }

    assert_checked_ranges(
        scrubber_db,
        [(swhids.ObjectType.SNAPSHOT, config_id)],
        {(swhids.ObjectType.SNAPSHOT, i, 16) for i in range(16)},
    )


def test_directory_duplicate_entries(scrubber_db, datastore, swh_storage):
    run_validators = attr.get_run_validators()
    attr.set_run_validators(False)
    try:
        invalid_directory = model.Directory(
            entries=(
                model.DirectoryEntry(
                    name=b"foo", type="dir", target=b"\x01" * 20, perms=1
                ),
                model.DirectoryEntry(
                    name=b"foo", type="file", target=b"\x00" * 20, perms=0
                ),
            )
        )
    finally:
        attr.set_run_validators(run_validators)
    swh_storage.directory_add([invalid_directory])

    deduplicated_directory = model.Directory(
        id=invalid_directory.id,
        entries=(
            model.DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
            model.DirectoryEntry(
                name=b"foo_0000000000", type="file", target=b"\x00" * 20, perms=0
            ),
        ),
        raw_manifest=(
            # fmt: off
            b"tree 52\x00"
            + b"0 foo\x00" + b"\x00" * 20
            + b"1 foo\x00" + b"\x01" * 20
            # fmt: on
        ),
    )

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add("cfg2", datastore, object_type, 1)
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == invalid_directory.swhid()
    assert corrupt_objects[0].config.datastore == datastore
    assert (
        before_date - datetime.timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + datetime.timedelta(seconds=5)
    )
    assert kafka_to_value(corrupt_objects[0].object_) == msgpack.loads(
        msgpack.dumps(deduplicated_directory.to_dict())  # turn entry list into tuple
    )

    assert_checked_ranges(
        scrubber_db, config, EXPECTED_PARTITIONS, before_date, after_date
    )


def test_no_recheck(scrubber_db, datastore, swh_storage):
    """
    Tests that objects that were already checked are not checked again on
    the next run.
    """
    # check the whole (empty) storage with a given config
    config_id = scrubber_db.config_add("cfg2", datastore, swhids.ObjectType.SNAPSHOT, 1)
    StorageChecker(
        db=scrubber_db,
        storage=swh_storage,
        config_id=config_id,
    ).run()

    # Corrupt two snapshots and add them to the storage
    snapshots = list(swh_model_data.SNAPSHOTS)
    for i in (0, 1):
        snapshots[i] = attr.evolve(snapshots[i], id=bytes([i]) * 20)
    swh_storage.snapshot_add(snapshots)

    previous_partitions = set(scrubber_db.checked_partition_iter(config_id))

    # rerun a checker for the same config, it should be a noop
    StorageChecker(
        db=scrubber_db,
        storage=swh_storage,
        config_id=config_id,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert (
        corrupt_objects == []
    ), "Detected corrupt objects in ranges that should have been skipped."

    # Make sure the DB was not changed (in particular, that timestamps were not bumped)
    assert set(scrubber_db.checked_partition_iter(config_id)) == previous_partitions


def test_no_hole(scrubber_db, datastore, swh_storage):
    swh_storage.content_add([CONTENT1])
    swh_storage.directory_add([DIRECTORY1, DIRECTORY2])
    swh_storage.revision_add([REVISION1])
    swh_storage.release_add([RELEASE1])
    swh_storage.snapshot_add([SNAPSHOT1])

    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add("cfg2", datastore, object_type, 1)
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()

    assert list(scrubber_db.missing_object_iter()) == []

    assert_checked_ranges(scrubber_db, config, EXPECTED_PARTITIONS)


@pytest.mark.parametrize(
    "missing_object",
    ["content1", "directory1", "directory2", "revision1", "release1"],
)
def test_one_hole(scrubber_db, datastore, swh_storage, missing_object):
    if missing_object == "content1":
        missing_swhid = CONTENT1.swhid()
        reference_swhids = [DIRECTORY1.swhid(), DIRECTORY2.swhid()]
    else:
        swh_storage.content_add([CONTENT1])

    if missing_object == "directory1":
        missing_swhid = DIRECTORY1.swhid()
        reference_swhids = [DIRECTORY2.swhid()]
    else:
        swh_storage.directory_add([DIRECTORY1])

    if missing_object == "directory2":
        missing_swhid = DIRECTORY2.swhid()
        reference_swhids = [REVISION1.swhid()]
    else:
        swh_storage.directory_add([DIRECTORY2])

    if missing_object == "revision1":
        missing_swhid = REVISION1.swhid()
        reference_swhids = [RELEASE1.swhid()]
    else:
        swh_storage.revision_add([REVISION1])

    if missing_object == "release1":
        missing_swhid = RELEASE1.swhid()
        reference_swhids = [SNAPSHOT1.swhid()]
    else:
        swh_storage.release_add([RELEASE1])

    swh_storage.snapshot_add([SNAPSHOT1])

    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add(
            f"cfg_{object_type.name}", datastore, object_type, 1
        )
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()

    assert [mo.id for mo in scrubber_db.missing_object_iter()] == [missing_swhid]
    assert {
        (mor.missing_id, mor.reference_id)
        for mor in scrubber_db.missing_object_reference_iter(missing_swhid)
    } == {(missing_swhid, reference_swhid) for reference_swhid in reference_swhids}

    assert_checked_ranges(scrubber_db, config, EXPECTED_PARTITIONS)


def test_two_holes(scrubber_db, datastore, swh_storage):
    # missing content and revision
    swh_storage.directory_add([DIRECTORY1, DIRECTORY2])
    swh_storage.release_add([RELEASE1])
    swh_storage.snapshot_add([SNAPSHOT1])

    config = []
    for object_type in OBJECT_TYPES:
        config_id = scrubber_db.config_add(
            f"cfg_{object_type.name}", datastore, object_type, 1
        )
        config.append((object_type, config_id))
        StorageChecker(
            db=scrubber_db,
            storage=swh_storage,
            config_id=config_id,
        ).run()

    assert {mo.id for mo in scrubber_db.missing_object_iter()} == {
        CONTENT1.swhid(),
        REVISION1.swhid(),
    }
    assert {
        mor.reference_id
        for mor in scrubber_db.missing_object_reference_iter(CONTENT1.swhid())
    } == {DIRECTORY1.swhid(), DIRECTORY2.swhid()}
    assert {
        mor.reference_id
        for mor in scrubber_db.missing_object_reference_iter(REVISION1.swhid())
    } == {RELEASE1.swhid()}

    assert_checked_ranges(scrubber_db, config, EXPECTED_PARTITIONS)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr
import pytest

from swh.journal.serializers import kafka_to_value
from swh.journal.writer import get_journal_writer
from swh.model import swhids
from swh.model.swhids import ObjectType
from swh.model.tests import swh_model_data
from swh.scrubber.db import Datastore
from swh.scrubber.journal_checker import JournalChecker, get_datastore


def journal_client_config(
    kafka_server: str, kafka_prefix: str, kafka_consumer_group: str
):
    return dict(
        cls="kafka",
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        on_eof="stop",
    )


def journal_writer(kafka_server: str, kafka_prefix: str):
    return get_journal_writer(
        cls="kafka",
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        anonymize=False,
    )


@pytest.fixture
def datastore(
    kafka_server: str, kafka_prefix: str, kafka_consumer_group: str
) -> Datastore:
    journal_config = journal_client_config(
        kafka_server, kafka_prefix, kafka_consumer_group
    )
    datastore = get_datastore(journal_config)
    return datastore


def test_no_corruption(
    scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group, datastore
):
    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("directory", swh_model_data.DIRECTORIES)
    writer.write_additions("revision", swh_model_data.REVISIONS)
    writer.write_additions("release", swh_model_data.RELEASES)
    writer.write_additions("snapshot", swh_model_data.SNAPSHOTS)

    journal_cfg = journal_client_config(
        kafka_server, kafka_prefix, kafka_consumer_group
    )
    gid = journal_cfg["group_id"] + "_"

    for object_type in ("directory", "revision", "release", "snapshot"):
        journal_cfg["group_id"] = gid + object_type
        config_id = scrubber_db.config_add(
            name=f"cfg_{object_type}",
            datastore=datastore,
            object_type=getattr(ObjectType, object_type.upper()),
            nb_partitions=1,
            check_references=False,
        )
        jc = JournalChecker(
            db=scrubber_db,
            config_id=config_id,
            journal=journal_cfg,
        )
        jc.run()
        jc.journal_client.close()

    assert list(scrubber_db.corrupt_object_iter()) == []


@pytest.mark.parametrize("corrupt_idx", range(len(swh_model_data.SNAPSHOTS)))
def test_corrupt_snapshot(
    scrubber_db,
    kafka_server,
    kafka_prefix,
    kafka_consumer_group,
    datastore,
    corrupt_idx,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=False,
    )
    snapshots = list(swh_model_data.SNAPSHOTS)
    snapshots[corrupt_idx] = attr.evolve(snapshots[corrupt_idx], id=b"\x00" * 20)

    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("snapshot", snapshots)

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    JournalChecker(
        db=scrubber_db,
        config_id=config_id,
        journal=journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group),
    ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == swhids.CoreSWHID.from_string(
        "swh:1:snp:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].config.datastore.package == "journal"
    assert corrupt_objects[0].config.datastore.cls == "kafka"
    assert (
        before_date - datetime.timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + datetime.timedelta(seconds=5)
    )
    assert (
        kafka_to_value(corrupt_objects[0].object_) == snapshots[corrupt_idx].to_dict()
    )


def test_corrupt_snapshots(
    scrubber_db,
    kafka_server,
    kafka_prefix,
    kafka_consumer_group,
    datastore,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=False,
    )
    snapshots = list(swh_model_data.SNAPSHOTS)
    for i in (0, 1):
        snapshots[i] = attr.evolve(snapshots[i], id=bytes([i]) * 20)

    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("snapshot", snapshots)

    JournalChecker(
        db=scrubber_db,
        config_id=config_id,
        journal=journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group),
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 2
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in [
            "swh:1:snp:0000000000000000000000000000000000000000",
            "swh:1:snp:0101010101010101010101010101010101010101",
        ]
    }


def test_check_references_raises(
    scrubber_db,
    kafka_server,
    kafka_prefix,
    kafka_consumer_group,
    datastore,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=True,
    )
    journal_config = journal_client_config(
        kafka_server, kafka_prefix, kafka_consumer_group
    )
    with pytest.raises(ValueError):
        JournalChecker(
            db=scrubber_db,
            config_id=config_id,
            journal=journal_config,
        )



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all known corrupts objects from the swh-scrubber database,
and tries to recover them.

Currently, only recovery from Git origins is implemented"""

import dataclasses
import functools
import logging
import os
from pathlib import Path
import subprocess
import tempfile
from typing import Dict, Optional, Type, Union

import dulwich
import dulwich.objects
import dulwich.repo
import psycopg2

from swh.journal.serializers import kafka_to_value, value_to_kafka
from swh.loader.git import converters
from swh.model.hashutil import hash_to_bytehex, hash_to_hex
from swh.model.model import BaseModel, Directory, Release, Revision, Snapshot
from swh.model.swhids import CoreSWHID, ObjectType

from .db import CorruptObject, FixedObject, ScrubberDb
from .utils import iter_corrupt_objects

logger = logging.getLogger(__name__)

ScrubbableObject = Union[Revision, Release, Snapshot, Directory]


def get_object_from_clone(
    clone_path: Path, swhid: CoreSWHID
) -> Union[None, bytes, dulwich.objects.ShaFile]:
    """Reads the original object matching the ``corrupt_object`` from the given clone
    if it exists, and returns a Dulwich object if possible, or a the raw manifest."""
    try:
        repo = dulwich.repo.Repo(str(clone_path))
    except dulwich.errors.NotGitRepository:
        return None

    with repo:  # needed to avoid packfile fd leaks
        try:
            return repo[hash_to_bytehex(swhid.object_id)]
        except KeyError:
            return None
        except dulwich.errors.ObjectFormatException:
            # fallback to git if dulwich can't parse it.
            # Unfortunately, Dulwich does not allow fetching an object without
            # parsing it into a ShaFile subclass, so we have to manually get it
            # by shelling out to git.
            object_type = (
                subprocess.check_output(
                    [
                        "git",
                        "-C",
                        clone_path,
                        "cat-file",
                        "-t",
                        hash_to_hex(swhid.object_id),
                    ]
                )
                .decode()
                .strip()
            )
            manifest = subprocess.check_output(
                [
                    "git",
                    "-C",
                    clone_path,
                    "cat-file",
                    object_type,
                    hash_to_hex(swhid.object_id),
                ]
            )
            manifest = f"{object_type} {len(manifest)}\x00".encode() + manifest
            logger.info("Dulwich failed to parse %r", manifest)
            return manifest


def get_fixed_object_from_clone(
    clone_path: Path, corrupt_object: CorruptObject
) -> Optional[FixedObject]:
    """Reads the original object matching the ``corrupt_object`` from the given clone
    if it exists, and returns a :class:`FixedObject` instance ready to be inserted
    in the database."""
    cloned_dulwich_obj_or_manifest = get_object_from_clone(
        clone_path, corrupt_object.id
    )
    if cloned_dulwich_obj_or_manifest is None:
        # Origin still exists, but object disappeared
        logger.info("%s not found in origin", corrupt_object.id)
        return None
    elif isinstance(cloned_dulwich_obj_or_manifest, bytes):
        # Dulwich could not parse it. Add as raw manifest to the existing object
        d = kafka_to_value(corrupt_object.object_)
        assert d.get("raw_manifest") is None, "Corrupt object has a raw_manifest"
        d["raw_manifest"] = cloned_dulwich_obj_or_manifest

        # Rebuild the object from the stored corrupt object + the raw manifest
        # just recovered; then checksum it.
        classes: Dict[ObjectType, Type[BaseModel]] = {
            ObjectType.REVISION: Revision,
            ObjectType.DIRECTORY: Directory,
            ObjectType.RELEASE: Release,
        }
        cls = classes[corrupt_object.id.object_type]
        recovered_obj = cls.from_dict(d)
        recovered_obj.check()

        return FixedObject(
            id=corrupt_object.id,
            object_=value_to_kafka(d),
            method="manifest_from_origin",
        )
    else:
        converter = {
            ObjectType.REVISION: converters.dulwich_commit_to_revision,
            ObjectType.DIRECTORY: converters.dulwich_tree_to_directory,
            ObjectType.RELEASE: converters.dulwich_tag_to_release,
        }[corrupt_object.id.object_type]
        cloned_obj = converter(cloned_dulwich_obj_or_manifest)

        # Check checksum, among others
        cloned_obj.check()

        return FixedObject(
            id=corrupt_object.id,
            object_=value_to_kafka(cloned_obj.to_dict()),
            method="from_origin",
        )


@dataclasses.dataclass
class Fixer:
    """Reads a chunk of corrupt objects in the swh-scrubber database, tries to recover
    them through various means (brute-forcing fields and re-downloading from the origin)
    recomputes checksums, and writes them back to the swh-scrubber database
    if successful.

    """

    db: ScrubberDb
    """Database to read from and write to."""
    start_object: CoreSWHID = CoreSWHID.from_string("swh:1:cnt:" + "00" * 20)
    """Minimum SWHID to check (in alphabetical order)"""
    end_object: CoreSWHID = CoreSWHID.from_string("swh:1:snp:" + "ff" * 20)
    """Maximum SWHID to check (in alphabetical order)"""

    def run(self):
        # TODO: currently only support re-downloading from the origin:
        # we should try brute-forcing for objects with no known origin (or when
        # all origins fail)
        after = ""
        while True:
            new_origins = self.db.object_origin_get(after=after)
            if not new_origins:
                break
            for origin_url in new_origins:
                self.recover_objects_from_origin(origin_url)
            after = new_origins[-1]

    def recover_objects_from_origin(self, origin_url):
        """Clones an origin, and cherry-picks original objects that are known to be
        corrupt in the database."""
        with tempfile.TemporaryDirectory(prefix=__name__ + ".") as tempdir:
            clone_path = Path(tempdir) / "repository.git"
            try:
                subprocess.run(
                    ["git", "clone", "--bare", origin_url, clone_path],
                    env={"PATH": os.environ["PATH"], "GIT_TERMINAL_PROMPT": "0"},
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    stdin=subprocess.DEVNULL,
                )
            except Exception:
                logger.exception("Failed to clone %s", origin_url)
                return

            iter_corrupt_objects(
                self.db,
                self.start_object,
                self.end_object,
                origin_url,
                functools.partial(self.recover_corrupt_object, clone_path=clone_path),
            )

    def recover_corrupt_object(
        self,
        corrupt_object: CorruptObject,
        cur: psycopg2.extensions.cursor,
        clone_path: Path,
    ) -> None:
        fixed_object = get_fixed_object_from_clone(clone_path, corrupt_object)

        if fixed_object is not None:
            self.db.fixed_object_add(cur, [fixed_object])



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from unittest.mock import MagicMock

import pytest

from swh.graph.http_naive_client import NaiveClient as NaiveGraphClient
from swh.model.model import Origin
from swh.model.swhids import CoreSWHID
from swh.scrubber.origin_locator import OriginLocator


@pytest.mark.parametrize("insert", [False, True])
def test_no_objects(scrubber_db, corrupt_object, insert):
    if insert:
        scrubber_db.corrupt_object_add(
            corrupt_object.id, corrupt_object.config, corrupt_object.object_
        )

    graph = MagicMock()
    storage = MagicMock()
    locator = OriginLocator(
        db=scrubber_db,
        graph=graph,
        storage=storage,
        # this range does not contain the object above
        start_object=CoreSWHID.from_string("swh:1:dir:00" + "00" * 19),
        end_object=CoreSWHID.from_string("swh:1:dir:60" + "00" * 19),
    )

    locator.run()

    assert graph.method_calls == []
    assert storage.method_calls == []

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM object_origin")
        assert cur.fetchone() == (0,)


def test_object_not_in_graph(scrubber_db, corrupt_object):
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )

    graph = NaiveGraphClient(nodes=[], edges=[])
    storage = MagicMock()
    locator = OriginLocator(
        db=scrubber_db,
        graph=graph,
        storage=storage,
        start_object=CoreSWHID.from_string("swh:1:dir:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:dir:" + "00" * 20),
    )

    locator.run()

    assert storage.method_calls == []

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM object_origin")
        assert cur.fetchone() == (0,)


def test_origin_not_in_storage(scrubber_db, swh_storage, corrupt_object, caplog):
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )

    origin = Origin(url="http://example.org")

    graph = NaiveGraphClient(
        nodes=[corrupt_object.id, origin.swhid()],
        edges=[(origin.swhid(), corrupt_object.id)],
    )
    locator = OriginLocator(
        db=scrubber_db,
        graph=graph,
        storage=swh_storage,
        start_object=CoreSWHID.from_string("swh:1:dir:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:dir:" + "ff" * 20),
    )

    with caplog.at_level(logging.ERROR, logger="swh.scrubber.origin_locator"):
        locator.run()

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM object_origin")
        assert cur.fetchone() == (0,)

    assert any(
        f"{origin.swhid()} found in graph but missing" in record[2]
        for record in caplog.record_tuples
    )


def test_two_origins(scrubber_db, corrupt_object, swh_storage):
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )

    origin1 = Origin(url="http://example.org")
    origin2 = Origin(url="http://example.com")
    swh_storage.origin_add([origin1, origin2])

    graph = NaiveGraphClient(
        nodes=[corrupt_object.id, origin1.swhid(), origin2.swhid()],
        edges=[
            (origin1.swhid(), corrupt_object.id),
            (origin2.swhid(), corrupt_object.id),
        ],
    )
    locator = OriginLocator(
        db=scrubber_db,
        graph=graph,
        storage=swh_storage,
        start_object=CoreSWHID.from_string("swh:1:dir:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:dir:" + "ff" * 20),
    )

    locator.run()

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT object_id, origin_url FROM object_origin")
        assert set(cur) == {
            (str(corrupt_object.id), origin1.url),
            (str(corrupt_object.id), origin2.url),
        }


def test_many_origins(scrubber_db, corrupt_object, swh_storage):
    scrubber_db.corrupt_object_add(
        corrupt_object.id, corrupt_object.config, corrupt_object.object_
    )

    origins = [Origin(url=f"http://example.org/{i}") for i in range(1000)]
    swh_storage.origin_add(origins)

    graph = NaiveGraphClient(
        nodes=[corrupt_object.id] + [origin.swhid() for origin in origins],
        edges=[(origin.swhid(), corrupt_object.id) for origin in origins],
    )
    locator = OriginLocator(
        db=scrubber_db,
        graph=graph,
        storage=swh_storage,
        start_object=CoreSWHID.from_string("swh:1:dir:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:dir:" + "ff" * 20),
    )

    locator.run()

    with scrubber_db.conn.cursor() as cur:
        cur.execute("SELECT object_id, origin_url FROM object_origin")
        rows = set(cur)
        assert rows <= {(str(corrupt_object.id), origin.url) for origin in origins}
        assert len(rows) == 100



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

import pytest

from swh.scrubber.db import Datastore

from .storage_checker_tests import *  # noqa


@pytest.fixture
def swh_storage_backend_config(swh_storage_cassandra_backend_config):
    return swh_storage_cassandra_backend_config


@pytest.fixture
def datastore(swh_storage):
    return Datastore(
        package="storage",
        cls="cassandra",
        instance=json.dumps(
            {
                "keyspace": swh_storage.keyspace,
                "hosts": swh_storage.hosts,
                "port": swh_storage.port,
            }
        ),
    )


@pytest.mark.skip(  # type: ignore[no-redef]
    "Duplicate directory entries are not representable in Cassandra, "
    "so this feature is irrelevant"
)
def test_directory_duplicate_entries():
    pass



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any

import pytest

from swh.scrubber import get_scrubber_db


@pytest.mark.parametrize("clz", ["local", "postgresql"])
def test_get_scrubber_db(mocker, clz):
    mock_scrubber = mocker.patch("swh.scrubber.db.ScrubberDb")

    def test_connect(db_str: str, **kwargs) -> Any:
        return "connection-result"

    mock_scrubber.connect.side_effect = test_connect

    actual_result = get_scrubber_db(clz, db="service=scrubber-db")

    assert mock_scrubber.connect.called is True
    assert actual_result == "connection-result"


@pytest.mark.parametrize("clz", ["something", "anything"])
def test_get_scrubber_db_raise(clz):
    assert clz not in ["local", "postgresql"]

    with pytest.raises(ValueError, match="Unknown"):
        get_scrubber_db(clz, db="service=scrubber-db")



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Callable, Optional

import psycopg2

from swh.model.swhids import CoreSWHID

from .db import CorruptObject, ScrubberDb


def iter_corrupt_objects(
    db: ScrubberDb,
    start_object: CoreSWHID,
    end_object: CoreSWHID,
    origin_url: Optional[str],
    cb: Callable[[CorruptObject, psycopg2.extensions.cursor], None],
) -> None:
    """Fetches objects and calls ``cb`` on each of them.

    objects are fetched with an update lock, with the same transaction as ``cb``,
    which is automatically committed after ``cb`` runs."""
    while True:
        with db.conn, db.cursor() as cur:
            if origin_url:
                corrupt_objects = list(
                    db.corrupt_object_grab_by_origin(
                        cur, origin_url, start_object, end_object
                    )
                )
            else:
                corrupt_objects = list(
                    db.corrupt_object_grab_by_id(cur, start_object, end_object)
                )
            if corrupt_objects and corrupt_objects[0].id == start_object:
                # TODO: don't needlessly fetch duplicate objects
                del corrupt_objects[0]
            if not corrupt_objects:
                # Nothing more to do
                break
            for corrupt_object in corrupt_objects:
                cb(corrupt_object, cur)
            db.conn.commit()  # XXX: is this redundant with db.conn.__exit__?

        start_object = corrupt_objects[-1].id



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import dataclasses
import datetime
import functools
from typing import Iterable, Iterator, List, Optional, Tuple

import psycopg2

from swh.core.db import BaseDb
from swh.model.swhids import CoreSWHID, ObjectType


def now():
    return datetime.datetime.now(tz=datetime.timezone.utc)


@dataclasses.dataclass(frozen=True)
class Datastore:
    """Represents a datastore being scrubbed; eg. swh-storage or swh-journal."""

    package: str
    """'storage', 'journal', or 'objstorage'."""
    cls: str
    """'postgresql'/'cassandra' for storage, 'kafka' for journal,
    'pathslicer'/'winery'/... for objstorage."""
    instance: str
    """Human readable string."""


@dataclasses.dataclass(frozen=True)
class ConfigEntry:
    """Represents a datastore being scrubbed; eg. swh-storage or swh-journal."""

    name: str
    datastore: Datastore
    object_type: ObjectType
    nb_partitions: int
    check_hashes: bool
    check_references: bool


@dataclasses.dataclass(frozen=True)
class CorruptObject:
    id: CoreSWHID
    config: ConfigEntry
    first_occurrence: datetime.datetime
    object_: bytes


@dataclasses.dataclass(frozen=True)
class MissingObject:
    id: CoreSWHID
    config: ConfigEntry
    first_occurrence: datetime.datetime


@dataclasses.dataclass(frozen=True)
class MissingObjectReference:
    missing_id: CoreSWHID
    reference_id: CoreSWHID
    config: ConfigEntry
    first_occurrence: datetime.datetime


@dataclasses.dataclass(frozen=True)
class FixedObject:
    id: CoreSWHID
    object_: bytes
    method: str
    recovery_date: Optional[datetime.datetime] = None


class ScrubberDb(BaseDb):
    current_version = 7

    ####################################
    # Shared tables
    ####################################

    @functools.lru_cache(1000)
    def datastore_get_or_add(self, datastore: Datastore) -> int:
        """Creates a datastore if it does not exist, and returns its id."""
        with self.transaction() as cur:
            cur.execute(
                """
                WITH inserted AS (
                    INSERT INTO datastore (package, class, instance)
                    VALUES (%(package)s, %(cls)s, %(instance)s)
                    ON CONFLICT DO NOTHING
                    RETURNING id
                )
                SELECT id
                FROM inserted
                UNION (
                    -- If the datastore already exists, we need to fetch its id
                    SELECT id
                    FROM datastore
                    WHERE
                        package=%(package)s
                        AND class=%(cls)s
                        AND instance=%(instance)s
                )
                LIMIT 1
                """,
                (dataclasses.asdict(datastore)),
            )
            res = cur.fetchone()
            assert res is not None
            (id_,) = res
            return id_

    @functools.lru_cache(1000)
    def datastore_get(self, datastore_id: int) -> Datastore:
        """Returns a datastore's id. Raises :exc:`ValueError` if it does not exist."""
        with self.transaction() as cur:
            cur.execute(
                """
                    SELECT package, class, instance
                    FROM datastore
                    WHERE id=%s
                """,
                (datastore_id,),
            )
            res = cur.fetchone()
            if res is None:
                raise ValueError(f"No datastore with id {datastore_id}")
            (package, cls, instance) = res
            return Datastore(package=package, cls=cls, instance=instance)

    def config_add(
        self,
        name: Optional[str],
        datastore: Datastore,
        object_type: ObjectType,
        nb_partitions: int,
        check_hashes: bool = True,
        check_references: bool = True,
    ) -> int:
        """Creates a configuration entry (and potentially a datastore);

        Will fail if a config with same (datastore. object_type, nb_paritions)
        already exists.
        """

        if not (check_hashes or check_references):
            raise ValueError(
                "At least one of the 2 check_hashes and check_references flags must be set"
            )
        datastore_id = self.datastore_get_or_add(datastore)
        if not name:
            name = (
                f"check_{object_type.name.lower()}_{nb_partitions}_"
                f"{datastore.package}_{datastore.cls}"
            )
        args = {
            "name": name,
            "datastore_id": datastore_id,
            "object_type": object_type.name.lower(),
            "nb_partitions": nb_partitions,
            "check_hashes": check_hashes,
            "check_references": check_references,
        }
        with self.transaction() as cur:
            cur.execute(
                """
                WITH inserted AS (
                    INSERT INTO check_config
                      (name, datastore, object_type, nb_partitions,
                       check_hashes, check_references)
                    VALUES
                      (%(name)s, %(datastore_id)s, %(object_type)s, %(nb_partitions)s,
                       %(check_hashes)s, %(check_references)s)
                    RETURNING id
                )
                SELECT id
                FROM inserted;
                """,
                args,
            )
            res = cur.fetchone()
            if res is None:
                raise ValueError(f"No config matching {args}")
            (id_,) = res
            return id_

    @functools.lru_cache(1000)
    def config_get(self, config_id: int) -> ConfigEntry:
        with self.transaction() as cur:
            cur.execute(
                """
                    SELECT
                      cc.name, cc.object_type, cc.nb_partitions,
                      cc.check_hashes, cc.check_references,
                      ds.package, ds.class, ds.instance
                    FROM check_config AS cc
                    INNER JOIN datastore As ds ON (cc.datastore=ds.id)
                    WHERE cc.id=%(config_id)s
                """,
                {
                    "config_id": config_id,
                },
            )
            res = cur.fetchone()
            if res is None:
                raise ValueError(f"No config with id {config_id}")
            (
                name,
                object_type,
                nb_partitions,
                chk_hashes,
                chk_refs,
                ds_package,
                ds_class,
                ds_instance,
            ) = res
            return ConfigEntry(
                name=name,
                datastore=Datastore(
                    package=ds_package, cls=ds_class, instance=ds_instance
                ),
                object_type=getattr(ObjectType, object_type.upper()),
                nb_partitions=nb_partitions,
                check_hashes=chk_hashes,
                check_references=chk_refs,
            )

    def config_get_by_name(
        self,
        name: str,
        datastore: Optional[int] = None,
    ) -> Optional[int]:
        """Get the configuration entry for given name, if any"""
        query_parts = ["SELECT id FROM check_config WHERE "]
        where_parts = [" name = %s "]
        query_params = [name]
        if datastore:
            where_parts.append(" datastore = %s ")
            query_params.append(str(datastore))

        query_parts.append(" AND ".join(where_parts))
        query = "\n".join(query_parts)
        with self.transaction() as cur:
            cur.execute(query, query_params)
            if cur.rowcount:
                res = cur.fetchone()
                if res:
                    (id_,) = res
                    return id_
            return None

    def config_iter(self) -> Iterator[Tuple[int, ConfigEntry]]:
        with self.transaction() as cur:
            cur.execute(
                """
                    SELECT
                      cc.id, cc.name, cc.object_type, cc.nb_partitions,
                      cc.check_hashes, cc.check_references,
                      ds.package, ds.class, ds.instance
                    FROM check_config AS cc
                    INNER JOIN datastore AS ds ON (cc.datastore=ds.id)
                """,
            )
            for row in cur:
                assert row is not None
                (
                    id_,
                    name,
                    object_type,
                    nb_partitions,
                    chk_hashes,
                    chk_refs,
                    ds_package,
                    ds_class,
                    ds_instance,
                ) = row
                yield (
                    id_,
                    ConfigEntry(
                        name=name,
                        datastore=Datastore(
                            package=ds_package, cls=ds_class, instance=ds_instance
                        ),
                        object_type=object_type,
                        nb_partitions=nb_partitions,
                        check_hashes=chk_hashes,
                        check_references=chk_refs,
                    ),
                )

    ####################################
    # Checkpointing/progress tracking
    ####################################

    def checked_partition_iter_next(
        self,
        config_id: int,
    ) -> Iterator[int]:
        """Generates partitions to be checked for the given configuration

        At each iteration, look for the next "free" partition in the
        checked_partition, for the given config_id, reserve it and return its
        id.

        Reserving the partition means make sure there is a row in the table for
        this partition id with the start_date column set.

        To chose a "free" partition is to select either the smaller partition
        is for which the start_date is NULL, or the first partition id not yet
        in the table.

        Stops the iteration when the number of partitions for the config id is
        reached.

        """
        while True:
            start_time = now()
            with self.transaction() as cur:
                cur.execute(
                    """
                    WITH next AS (
                     SELECT min(partition_id) as pid
                      FROM checked_partition
                      WHERE config_id=%(config_id)s and start_date is NULL
                     UNION
                     SELECT coalesce(max(partition_id) + 1, 0) as pid
                      FROM checked_partition
                      WHERE config_id=%(config_id)s
                    )
                    INSERT INTO checked_partition(config_id, partition_id, start_date)
                      select %(config_id)s, min(pid), %(start_date)s from next
                      where pid is not NULL
                    ON CONFLICT (config_id, partition_id)
                    DO UPDATE
                      SET start_date = GREATEST(
                        checked_partition.start_date, EXCLUDED.start_date
                      )
                    RETURNING partition_id;
                    """,
                    {"config_id": config_id, "start_date": start_time},
                )
                res = cur.fetchone()
                assert res is not None
                (partition_id,) = res
                if partition_id >= self.config_get(config_id).nb_partitions:
                    self.conn.rollback()
                    return
            yield partition_id

    def checked_partition_reset(self, config_id: int, partition_id: int) -> bool:
        """
        Reset the partition, aka clear start_date and end_date
        """
        with self.transaction() as cur:
            cur.execute(
                """
                UPDATE checked_partition
                SET start_date=NULL, end_date=NULL
                WHERE config_id=%(config_id)s AND partition_id=%(partition_id)s
                """,
                {"config_id": config_id, "partition_id": partition_id},
            )
            return bool(cur.rowcount)

    def checked_partition_upsert(
        self,
        config_id: int,
        partition_id: int,
        date: Optional[datetime.datetime] = None,
    ) -> None:
        """
        Records in the database the given partition was last checked at the given date.
        """
        if date is None:
            date = now()

        with self.transaction() as cur:
            cur.execute(
                """
                UPDATE checked_partition
                SET end_date = GREATEST(%(date)s, end_date)
                WHERE config_id=%(config_id)s AND partition_id=%(partition_id)s
                """,
                {
                    "config_id": config_id,
                    "partition_id": partition_id,
                    "date": date,
                },
            )

    def checked_partition_get_last_date(
        self,
        config_id: int,
        partition_id: int,
    ) -> Optional[datetime.datetime]:
        """
        Returns the last date the given partition was checked in the given datastore,
        or :const:`None` if it was never checked.

        Currently, this matches partition id and number exactly, with no regard for
        partitions that contain or are contained by it.
        """
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT end_date
                FROM checked_partition
                WHERE config_id=%s AND partition_id=%s
                """,
                (config_id, partition_id),
            )

            res = cur.fetchone()
            if res is None:
                return None
            else:
                (date,) = res
                return date

    def checked_partition_get_running(
        self,
        config_id: int,
    ) -> Iterator[Tuple[int, datetime.datetime]]:
        """Yields the partitions which are currently being checked; i.e. which have a
        start_date but no end_date.
        """
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT partition_id, start_date
                FROM checked_partition
                WHERE config_id=%s AND end_date is NULL
                """,
                (config_id,),
            )

            for partition_id, start_date in cur:
                yield (partition_id, start_date)

    def checked_partition_get_stuck(
        self,
        config_id: int,
        since: Optional[datetime.timedelta] = None,
    ) -> Iterator[Tuple[int, datetime.datetime]]:
        """Yields the partitions which are currently running for more than `since`; if
        not set, automatically guess a reasonable delay from completed partitions.
        If no such a delay can be extracted, fall back to 1 hour.

        The heuristic for the automatic delay is 2x max(end_date-start_date)
        for the last 10 partitions checked.

        """
        with self.transaction() as cur:
            if since is None:
                cur.execute(
                    """
                    WITH delays as
                    (
                    SELECT end_date - start_date as delay
                    FROM checked_partition
                    WHERE config_id=%s AND end_date is not NULL
                    ORDER BY start_date DESC
                    LIMIT 10
                    )
                    SELECT 2*max(delay) from delays
                    """,
                    (config_id,),
                )
                res = cur.fetchone()
                assert res is not None
                (since,) = res
            if since is None:
                since = datetime.timedelta(hours=1)

            cur.execute(
                """
                SELECT partition_id, start_date
                FROM checked_partition
                WHERE config_id=%s AND end_date is NULL AND start_date < %s
                """,
                (config_id, now() - since),
            )

            for partition_id, start_date in cur:
                yield (partition_id, start_date)

    def checked_partition_iter(
        self, config_id: int
    ) -> Iterator[Tuple[int, int, datetime.datetime, Optional[datetime.datetime]]]:
        """Yields tuples of ``(partition_id, nb_partitions, start_date, end_date)``"""
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT CP.partition_id, CC.nb_partitions, CP.start_date, CP.end_date
                FROM checked_partition as CP
                INNER JOIN check_config AS CC on (CC.id=CP.config_id)
                WHERE CC.id=%s
                """,
                (config_id,),
            )

            for row in cur:
                yield tuple(row)  # type: ignore[misc]

    ####################################
    # Inventory of objects with issues
    ####################################

    def corrupt_object_add(
        self,
        id: CoreSWHID,
        config: ConfigEntry,
        serialized_object: bytes,
    ) -> None:
        config_id = self.config_get_by_name(config.name)
        assert config_id is not None
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO corrupt_object (id, config_id, object)
                VALUES (%s, %s, %s)
                ON CONFLICT DO NOTHING
                """,
                (str(id), config_id, serialized_object),
            )

    def _corrupt_object_list_from_cursor(
        self, cur: psycopg2.extensions.cursor
    ) -> Iterator[CorruptObject]:
        for row in cur:
            (
                id,
                first_occurrence,
                object_,
                cc_object_type,
                cc_nb_partitions,
                cc_name,
                cc_chk_hashes,
                cc_chk_refs,
                ds_package,
                ds_class,
                ds_instance,
            ) = row
            yield CorruptObject(
                id=CoreSWHID.from_string(id),
                first_occurrence=first_occurrence,
                object_=object_,
                config=ConfigEntry(
                    name=cc_name,
                    datastore=Datastore(
                        package=ds_package, cls=ds_class, instance=ds_instance
                    ),
                    object_type=cc_object_type,
                    nb_partitions=cc_nb_partitions,
                    check_hashes=cc_chk_hashes,
                    check_references=cc_chk_refs,
                ),
            )

    def corrupt_object_iter(self) -> Iterator[CorruptObject]:
        """Yields all records in the 'corrupt_object' table."""
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT
                    co.id, co.first_occurrence, co.object,
                    cc.object_type, cc.nb_partitions, cc.name,
                    cc.check_hashes, cc.check_references,
                    ds.package, ds.class, ds.instance
                FROM corrupt_object AS co
                INNER JOIN check_config AS cc ON (cc.id=co.config_id)
                INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
                """
            )
            yield from self._corrupt_object_list_from_cursor(cur)

    def corrupt_object_get(
        self,
        start_id: CoreSWHID,
        end_id: CoreSWHID,
        limit: int = 100,
    ) -> Iterator[CorruptObject]:
        """Yields a page of records in the 'corrupt_object' table, ordered by id.

        Arguments:
            start_id: Only return objects after this id
            end_id: Only return objects before this id
            in_origin: An origin URL. If provided, only returns objects that may be
                found in the given origin
        """
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT
                    co.id, co.first_occurrence, co.object,
                    cc.object_type, cc.nb_partitions, cc.name,
                    cc.check_hashes, cc.check_references,
                    ds.package, ds.class, ds.instance
                FROM corrupt_object AS co
                INNER JOIN check_config AS cc ON (cc.id=co.config_id)
                INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
                WHERE
                    co.id >= %s
                    AND co.id <= %s
                ORDER BY co.id
                LIMIT %s
                """,
                (str(start_id), str(end_id), limit),
            )
            yield from self._corrupt_object_list_from_cursor(cur)

    def corrupt_object_grab_by_id(
        self,
        cur: psycopg2.extensions.cursor,
        start_id: CoreSWHID,
        end_id: CoreSWHID,
        limit: int = 100,
    ) -> Iterator[CorruptObject]:
        """Returns a page of records in the 'corrupt_object' table for a fixer,
        ordered by id

        These records are not already fixed (ie. do not have a corresponding entry
        in the 'fixed_object' table), and they are selected with an exclusive update
        lock.

        Arguments:
            start_id: Only return objects after this id
            end_id: Only return objects before this id
        """
        cur.execute(
            """
            SELECT
                co.id, co.first_occurrence, co.object,
                cc.object_type, cc.nb_partitions, cc.name,
                cc.check_hashes, cc.check_references,
                ds.package, ds.class, ds.instance
            FROM corrupt_object AS co
            INNER JOIN check_config AS cc ON (cc.id=co.config_id)
            INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
            WHERE
                co.id >= %(start_id)s
                AND co.id <= %(end_id)s
                AND NOT EXISTS (SELECT 1 FROM fixed_object WHERE fixed_object.id=co.id)
            ORDER BY co.id
            LIMIT %(limit)s
            FOR UPDATE SKIP LOCKED
            """,
            dict(
                start_id=str(start_id),
                end_id=str(end_id),
                limit=limit,
            ),
        )
        yield from self._corrupt_object_list_from_cursor(cur)

    def corrupt_object_grab_by_origin(
        self,
        cur: psycopg2.extensions.cursor,
        origin_url: str,
        start_id: Optional[CoreSWHID] = None,
        end_id: Optional[CoreSWHID] = None,
        limit: int = 100,
    ) -> Iterator[CorruptObject]:
        """Returns a page of records in the 'corrupt_object' table for a fixer,
        ordered by id

        These records are not already fixed (ie. do not have a corresponding entry
        in the 'fixed_object' table), and they are selected with an exclusive update
        lock.

        Arguments:
            origin_url: only returns objects that may be found in the given origin
        """
        cur.execute(
            """
            SELECT
                co.id, co.first_occurrence, co.object,
                cc.object_type, cc.nb_partitions, cc.name,
                cc.check_hashes, cc.check_references,
                ds.package, ds.class, ds.instance
            FROM corrupt_object AS co
            INNER JOIN check_config AS cc ON (cc.id=co.config_id)
            INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
            INNER JOIN object_origin AS oo ON (oo.object_id=co.id)
            WHERE
                (co.id >= %(start_id)s OR %(start_id)s IS NULL)
                AND (co.id <= %(end_id)s OR %(end_id)s IS NULL)
                AND NOT EXISTS (SELECT 1 FROM fixed_object WHERE fixed_object.id=co.id)
                AND oo.origin_url=%(origin_url)s
            ORDER BY co.id
            LIMIT %(limit)s
            FOR UPDATE SKIP LOCKED
            """,
            dict(
                start_id=None if start_id is None else str(start_id),
                end_id=None if end_id is None else str(end_id),
                origin_url=origin_url,
                limit=limit,
            ),
        )
        yield from self._corrupt_object_list_from_cursor(cur)

    def missing_object_add(
        self,
        id: CoreSWHID,
        reference_ids: Iterable[CoreSWHID],
        config: ConfigEntry,
    ) -> None:
        """
        Adds a "hole" to the inventory, ie. an object missing from a datastore
        that is referenced by an other object of the same datastore.

        If the missing object is already known to be missing by the scrubber database,
        this only records the reference (which can be useful to locate an origin
        to recover the object from).
        If that reference is already known too, this is a noop.

        Args:
            id: SWHID of the missing object (the hole)
            reference_id: SWHID of the object referencing the missing object
            datastore: representation of the swh-storage/swh-journal/... instance
              containing this hole
        """
        if not reference_ids:
            raise ValueError("reference_ids is empty")
        config_id = self.config_get_by_name(config.name)
        with self.transaction() as cur:
            cur.execute(
                """
                INSERT INTO missing_object (id, config_id)
                VALUES (%s, %s)
                ON CONFLICT DO NOTHING
                """,
                (str(id), config_id),
            )
            psycopg2.extras.execute_batch(
                cur,
                """
                INSERT INTO missing_object_reference (missing_id, reference_id, config_id)
                VALUES (%s, %s, %s)
                ON CONFLICT DO NOTHING
                """,
                [
                    (str(id), str(reference_id), config_id)
                    for reference_id in reference_ids
                ],
            )

    def missing_object_iter(self) -> Iterator[MissingObject]:
        """Yields all records in the 'missing_object' table."""
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT
                    mo.id, mo.first_occurrence,
                    cc.name, cc.object_type, cc.nb_partitions,
                    cc.check_hashes, cc.check_references,
                    ds.package, ds.class, ds.instance
                FROM missing_object AS mo
                INNER JOIN check_config AS cc ON (cc.id=mo.config_id)
                INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
                """
            )

            for row in cur:
                (
                    id,
                    first_occurrence,
                    cc_name,
                    cc_object_type,
                    cc_nb_partitions,
                    cc_chk_hashes,
                    cc_chk_refs,
                    ds_package,
                    ds_class,
                    ds_instance,
                ) = row
                yield MissingObject(
                    id=CoreSWHID.from_string(id),
                    first_occurrence=first_occurrence,
                    config=ConfigEntry(
                        name=cc_name,
                        object_type=cc_object_type,
                        nb_partitions=cc_nb_partitions,
                        check_hashes=cc_chk_hashes,
                        check_references=cc_chk_refs,
                        datastore=Datastore(
                            package=ds_package, cls=ds_class, instance=ds_instance
                        ),
                    ),
                )

    def missing_object_reference_iter(
        self, missing_id: CoreSWHID
    ) -> Iterator[MissingObjectReference]:
        """Yields all records in the 'missing_object_reference' table."""
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT
                    mor.reference_id, mor.first_occurrence,
                    cc.name, cc.object_type, cc.nb_partitions,
                    cc.check_hashes, cc.check_references,
                    ds.package, ds.class, ds.instance
                FROM missing_object_reference AS mor
                INNER JOIN check_config AS cc ON (cc.id=mor.config_id)
                INNER JOIN datastore AS ds ON (ds.id=cc.datastore)
                WHERE mor.missing_id=%s
                """,
                (str(missing_id),),
            )

            for row in cur:
                (
                    reference_id,
                    first_occurrence,
                    cc_name,
                    cc_object_type,
                    cc_nb_partitions,
                    cc_chk_hashes,
                    cc_chk_refs,
                    ds_package,
                    ds_class,
                    ds_instance,
                ) = row
                yield MissingObjectReference(
                    missing_id=missing_id,
                    reference_id=CoreSWHID.from_string(reference_id),
                    first_occurrence=first_occurrence,
                    config=ConfigEntry(
                        name=cc_name,
                        object_type=cc_object_type,
                        nb_partitions=cc_nb_partitions,
                        check_hashes=cc_chk_hashes,
                        check_references=cc_chk_refs,
                        datastore=Datastore(
                            package=ds_package, cls=ds_class, instance=ds_instance
                        ),
                    ),
                )

    ####################################
    # Issue resolution
    ####################################

    def object_origin_add(
        self, cur: psycopg2.extensions.cursor, swhid: CoreSWHID, origins: List[str]
    ) -> None:
        psycopg2.extras.execute_values(
            cur,
            """
            INSERT INTO object_origin (object_id, origin_url)
            VALUES %s
            ON CONFLICT DO NOTHING
            """,
            [(str(swhid), origin_url) for origin_url in origins],
        )

    def object_origin_get(self, after: str = "", limit: int = 1000) -> List[str]:
        """Returns origins with non-fixed corrupt objects, ordered by URL.

        Arguments:
            after: if given, only returns origins with an URL after this value
        """
        with self.transaction() as cur:
            cur.execute(
                """
                SELECT DISTINCT origin_url
                FROM object_origin
                WHERE
                    origin_url > %(after)s
                    AND object_id IN (
                        (SELECT id FROM corrupt_object)
                        EXCEPT (SELECT id FROM fixed_object)
                    )
                ORDER BY origin_url
                LIMIT %(limit)s
                """,
                dict(after=after, limit=limit),
            )

            return [origin_url for (origin_url,) in cur]

    def fixed_object_add(
        self, cur: psycopg2.extensions.cursor, fixed_objects: List[FixedObject]
    ) -> None:
        psycopg2.extras.execute_values(
            cur,
            """
            INSERT INTO fixed_object (id, object, method)
            VALUES %s
            ON CONFLICT DO NOTHING
            """,
            [
                (str(fixed_object.id), fixed_object.object_, fixed_object.method)
                for fixed_object in fixed_objects
            ],
        )

    def fixed_object_iter(self) -> Iterator[FixedObject]:
        with self.transaction() as cur:
            cur.execute("SELECT id, object, method, recovery_date FROM fixed_object")
            for (id, object_, method, recovery_date) in cur:
                yield FixedObject(
                    id=CoreSWHID.from_string(id),
                    object_=object_,
                    method=method,
                    recovery_date=recovery_date,
                )



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all objects in a swh-storage instance and recomputes their checksums."""

import json
import logging
from typing import Any, Dict, List, Optional

from swh.journal.client import get_journal_client
from swh.journal.serializers import kafka_to_value
from swh.model import model

from .db import ConfigEntry, Datastore, ScrubberDb

logger = logging.getLogger(__name__)


def get_datastore(journal_cfg) -> Datastore:
    if journal_cfg.get("cls") == "kafka":
        datastore = Datastore(
            package="journal",
            cls="kafka",
            instance=json.dumps(
                {
                    "brokers": journal_cfg["brokers"],
                    "group_id": journal_cfg["group_id"],
                    "prefix": journal_cfg["prefix"],
                }
            ),
        )
    else:
        raise NotImplementedError(
            f"JournalChecker(journal={journal_cfg!r}).datastore()"
        )
    return datastore


class JournalChecker:
    """Reads a chunk of a swh-storage database, recomputes checksums, and
    reports errors in a separate database."""

    _config: Optional[ConfigEntry] = None
    _datastore: Optional[Datastore] = None

    def __init__(self, db: ScrubberDb, config_id: int, journal: Dict[str, Any]):
        self.db = db
        self.config_id = config_id

        if self.config.check_references:
            raise ValueError(
                "The journal checcker cannot check for references, please set "
                "the 'check_references' to False in the config entry %s.",
                self.config_id,
            )
        self.journal_client_config = journal.copy()
        if "object_types" in self.journal_client_config:
            raise ValueError(
                "The journal_client configuration entry should not define the "
                "object_types field; this is handled by the scrubber configuration entry"
            )
        self.journal_client_config["object_types"] = [
            self.config.object_type.name.lower()
        ]
        self.journal_client = get_journal_client(
            **self.journal_client_config,
            # Remove default deserializer; so process_kafka_values() gets the message
            # verbatim so it can archive it with as few modifications a possible.
            value_deserializer=lambda obj_type, msg: msg,
        )

    @property
    def config(self) -> ConfigEntry:
        if self._config is None:
            self._config = self.db.config_get(self.config_id)

        assert self._config is not None
        return self._config

    @property
    def datastore(self) -> Datastore:
        """Returns a :class:`Datastore` instance representing the journal instance
        being checked."""
        return self.config.datastore

    def run(self):
        """Runs a journal client with the given configuration.
        This method does not return, unless otherwise configured (with ``stop_on_eof``).
        """
        self.journal_client.process(self.process_kafka_messages)

    def process_kafka_messages(self, all_messages: Dict[str, List[bytes]]):
        for (object_type, messages) in all_messages.items():
            logger.debug("Processing %s %s", len(messages), object_type)
            cls = getattr(model, object_type.capitalize())
            for message in messages:
                object_ = cls.from_dict(kafka_to_value(message))
                real_id = object_.compute_hash()
                if object_.id != real_id:
                    self.db.corrupt_object_add(object_.swhid(), self.config, message)



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import os
from subprocess import CalledProcessError

from click.testing import CliRunner
import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import import_swhmodule, swh_db_upgrade, swh_db_version
from swh.core.db.tests.test_cli import craft_conninfo
from swh.model.swhids import ObjectType
from swh.scrubber.db import ScrubberDb


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture()
def current_version():
    return 6


def test_datadir(datadir):
    assert datadir.endswith("/swh/scrubber/tests/data")


@pytest.fixture()
def mock_import_swhmodule(mocker, datadir, current_version):
    """This bypasses the module manipulation to make import_swhmodule return a mock
    object suitable for data test files listing via get_sql_for_package.

    For a given module `test.<mod>`, return a MagicMock object with a __name__
    set to `<mod>` and __file__ pointing to `data/<mod>/__init__.py`.

    The Mock object also defines a `get_datastore()` attribute on which the
    `current_version` attribute is set to `current_version`.

    Typical usage::

      def test_xxx(cli_runner, mock_import_swhmodule):
        conninfo = craft_conninfo(test_db, "new-db")
        module_name = "test.cli"
        # the command below will use sql scripts from
        #     swh/core/db/tests/data/cli/sql/*.sql
        cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])

    """

    def import_swhmodule_mock(modname):
        if modname.startswith("test."):
            dirname = modname.split(".", 1)[1]

            def get_datastore(*args, **kw):
                return mocker.MagicMock(current_version=current_version)

            return mocker.MagicMock(
                __name__=modname,
                __file__=os.path.join(datadir, dirname, "__init__.py"),
                get_datastore=get_datastore,
            )
        else:
            return import_swhmodule(modname)

    return mocker.patch("swh.core.db.db_utils.import_swhmodule", import_swhmodule_mock)


def test_upgrade_6_to_7(
    cli_runner, postgresql, mock_import_swhmodule, datadir, current_version
):
    """Check 6 to 7 migration"""
    module = "test.cli"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_version(conninfo) == current_version

    new_version = swh_db_upgrade(conninfo, module, 7)
    assert new_version == 7
    assert swh_db_version(conninfo) == 7

    tmap = {otype.name.lower(): otype.value for otype in ObjectType}
    db = ScrubberDb(postgresql)

    # corrupt objects
    corrupt_objects = list(db.corrupt_object_iter())
    assert len(corrupt_objects) == 15

    got = [
        (
            o.config.datastore.instance,
            o.config.object_type,
            o.config.name,
            str(o.id),
        )
        for o in corrupt_objects
    ]

    # expected result for corrupt objects
    expected = []
    for otype, n, ds, port in (
        ("release", 16, 1, 5432),
        ("revision", 16, 1, 5432),
        ("release", 16, 2, 5433),
    ):
        expected += [
            (str(port), otype, f"ds{ds}-{otype}-{n}", f"swh:1:{tmap[otype]}:{i:040X}")
            for i in range(5)
        ]

    assert set(got) == set(expected)

    # missing objects
    missing_objects = list(db.missing_object_iter())
    assert len(missing_objects) == 15

    got = [
        (o.config.datastore.instance, o.config.object_type, o.config.name, str(o.id))
        for o in missing_objects
    ]

    # expected result for missing objects
    # For those, the object type of the check_config entry is the one of the
    # reference object (not the one of the missing object).
    # E.g. the check_config for snapshot generates release missing object entries.
    expected = []
    for checked_otype, missing_otype, n, ds, port in (
        ("snapshot", "release", 16, 1, 5432),
        ("snapshot", "revision", 16, 1, 5432),
        ("snapshot", "release", 16, 2, 5433),
    ):
        expected += [
            (
                str(port),
                checked_otype,
                f"ds{ds}-{checked_otype}-{n}",
                f"swh:1:{tmap[missing_otype]}:{i:040X}",
            )
            for i in range(5)
        ]

    assert set(got) == set(expected)

    # missing_object_references
    for mo in missing_objects:
        mo_refs = list(db.missing_object_reference_iter(mo.id))
        assert mo.config in [mor.config for mor in mo_refs]
        # for each missing object, there is only one reference from the same
        # check session (same check_config)
        assert len([mor for mor in mo_refs if mor.config == mo.config]) == 1


def test_upgrade_6_to_7_fails_corrupt(
    cli_runner, postgresql, mock_import_swhmodule, datadir, current_version
):
    """Check 6 to 7 migration fails

    in case there is a corrupt_object row with a datastore that matches 2
    check_configs for the object type

    """

    module = "test.cli"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_version(conninfo) == current_version

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        # datastore 3 have 2 check_config entries for the release object_type
        cur.execute(
            "insert into corrupt_object(id, datastore, object) " "values (%s, %s, %s)",
            ("swh:1:rel:0000000000000000000000000000000000000000", 3, b"\x00"),
        )
    with pytest.raises(CalledProcessError):
        swh_db_upgrade(conninfo, module, 7)

    assert swh_db_version(conninfo) == 6


def test_upgrade_6_to_7_fails_missing_reference(
    cli_runner, postgresql, mock_import_swhmodule, datadir, current_version
):
    """Check 6 to 7 migration fails

    in case there is a missing_object_reference row with a datastore that matches 2
    check_configs for the object (reference_id) type.
    """
    module = "test.cli"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_version(conninfo) == current_version

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        # datastore 3 have 2 check_config entries for the release object_type
        cur.execute(
            "insert into missing_object_reference(missing_id, reference_id, datastore) "
            "values (%s, %s, %s)",
            (
                "swh:1:dir:0000000000000000000000000000000000000000",
                "swh:1:rel:0000000000000000000000000000000000000000",
                3,
            ),
        )

    with pytest.raises(CalledProcessError):
        swh_db_upgrade(conninfo, module, 7)

    assert swh_db_version(conninfo) == 6


def test_upgrade_6_to_7_fails_missing(
    cli_runner, postgresql, mock_import_swhmodule, datadir, current_version
):
    """Check 6 to 7 migration fails

    in case there is a missing_object row with a datastore that matches 2
    check_configs for the object type.
    """

    module = "test.cli"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_version(conninfo) == current_version

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        # datastore 3 have 2 check_config entries for the release object_type
        cur.execute(
            "insert into missing_object(id, datastore) " "values (%s, %s)",
            ("swh:1:rel:0000000000000000000000000000000000000000", 3),
        )

    with pytest.raises(CalledProcessError):
        swh_db_upgrade(conninfo, module, 7)

    assert swh_db_version(conninfo) == 6



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from functools import partial

import attr
import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.journal.serializers import value_to_kafka
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Directory, DirectoryEntry
from swh.model.swhids import ObjectType
from swh.model.tests.swh_model_data import DIRECTORIES
from swh.scrubber.db import ConfigEntry, CorruptObject, Datastore, ScrubberDb

scrubber_postgresql_proc = factories.postgresql_proc(
    load=[partial(initialize_database_for_module, modname="scrubber", version=6)],
)

postgresql_scrubber = factories.postgresql("scrubber_postgresql_proc")

OBJECT_TYPE = ObjectType.DIRECTORY
PARTITION_ID = 2
NB_PARTITIONS = 64
DIRECTORY = [dir_ for dir_ in DIRECTORIES if len(dir_.entries) > 1][0]
ORIGINAL_DIRECTORY = Directory(
    entries=(
        DirectoryEntry(
            name=b"dir1",
            type="dir",
            target=hash_to_bytes("4b825dc642cb6eb9a060e54bf8d69288fbee4904"),
            perms=0o040755,
        ),
        DirectoryEntry(
            name=b"file1.ext",
            type="file",
            target=hash_to_bytes("86bc6b377e9d25f9d26777a4a28d08e63e7c5779"),
            perms=0o644,
        ),
        DirectoryEntry(
            name=b"subprepo1",
            type="rev",
            target=hash_to_bytes("c7f96242d73c267adc77c2908e64e0c1cb6a4431"),
            perms=0o160000,
        ),
    ),
    raw_manifest=(
        b"tree 102\x00"
        b"160000 subprepo1\x00\xc7\xf9bB\xd7<&z\xdcw\xc2\x90\x8ed\xe0\xc1\xcbjD1"
        b"644 file1.ext\x00\x86\xbck7~\x9d%\xf9\xd2gw\xa4\xa2\x8d\x08\xe6>|Wy"
        b"40755 dir1\x00K\x82]\xc6B\xcbn\xb9\xa0`\xe5K\xf8\xd6\x92\x88\xfb\xeeI\x04"
    ),
)

# A directory with its entries in canonical order, but a hash computed as if
# computed in the reverse order.
# This happens when entries get normalized (either by the loader or accidentally
# in swh-storage)
CORRUPT_DIRECTORY = attr.evolve(ORIGINAL_DIRECTORY, raw_manifest=None)
assert ORIGINAL_DIRECTORY != CORRUPT_DIRECTORY
assert (
    hash_to_bytes("61992617462fff81509bda4a24b54c96ea74a007")
    == ORIGINAL_DIRECTORY.id
    == CORRUPT_DIRECTORY.id
)
assert (
    hash_to_bytes("81fda5b242e65fc81201e590d0f0ce5f582fbcdd")
    == CORRUPT_DIRECTORY.compute_hash()
    != CORRUPT_DIRECTORY.id
)
assert ORIGINAL_DIRECTORY.entries == CORRUPT_DIRECTORY.entries


@pytest.fixture
def datastore():
    return Datastore(package="storage", cls="postgresql", instance="service=swh-test")


@pytest.fixture
def scrubber_db(postgresql_scrubber):
    db = ScrubberDb(postgresql_scrubber)
    with db.conn.cursor() as cur:
        cur.execute("TRUNCATE TABLE corrupt_object")
        cur.execute("TRUNCATE TABLE datastore CASCADE")
    return db


@pytest.fixture
def config_id(scrubber_db, datastore) -> int:
    return scrubber_db.config_add(
        f"cfg_{OBJECT_TYPE}_{NB_PARTITIONS}", datastore, OBJECT_TYPE, NB_PARTITIONS
    )


@pytest.fixture
def config_entry(scrubber_db, config_id) -> ConfigEntry:
    return scrubber_db.config_get(config_id)


@pytest.fixture
def corrupt_object(scrubber_db, config_entry):
    return CorruptObject(
        id=ORIGINAL_DIRECTORY.swhid(),
        config=config_entry,
        first_occurrence=datetime.datetime.now(tz=datetime.timezone.utc),
        object_=value_to_kafka(CORRUPT_DIRECTORY.to_dict()),
    )
